//Part of XDBEngine source code.
//Version 1.4
//Created by Tomas Koutny.
//for more info see http://www.rawos.cz/xdbe/index.html
//This file is public domain.


//interface for ordinary user to manipulate with databases
unit UserLevel;

interface

uses UserStructs, windows;

type
    TXDBase = class;
    TXDBTable = class;
    TXDBColumn = class;
    TXDBDataObject = class;

    TXDBEngine = class   //cache & db files owner & creator
                 protected

                   FYesStr,
                   FNoStr:WideString;//support for text representation of Yes/No

                   FLastError:integer;    //last error
                   FCacheSize:LongWord; //size of memory allocated  for cache
                   function GetFilter(index:integer):TFilterInfo; virtual; abstract;
                            //gets filter's info
                 public
                   function OpenDBase(DBName:String; CreateIt:Boolean;
                            FilterID:integer; var handle:TXDBase):integer; dynamic; abstract;
                             //creates or open new database
                             //FilterID determines kind of this databse
                             //handle is newly created or opened database
                             //returns constants defined in XDBRetCodes
                             //if file DBName doesn't exists then CreatIt:=True

                   function FiltersCount:integer; virtual; abstract; //returns
                            //count of installed filters
                   function FilterIOCtl(index:integer; ACode:integer;
                                      AData:pointer):integer; virtual; abstract;
                            //executes TFilter.IOCtl(ACode, AData)

                   property Filters[index:integer]:TFilterInfo read GetFilter; default;
                   property LastError:Integer read FLastError;

                   property YesString:WideString read FYesStr write FYesStr;
                   property NoString:WideString read FNoStr write FNoStr;

                   property CacheSize:LongWord read FCacheSize;//current size
                            //of memory allocated for cache
                 end;

    TXDBase = class      //object database representation
              protected
                function GetTable(index:integer):TXDBTable; virtual; abstract;
                function GetNamedTable(TblName:WideString):TXDBTable; virtual; abstract;
              public
                 procedure Flush; virtual; abstract; //saves data to disk
                 procedure Assimilate(Who:TXDBase); virtual; abstract;
                        //imports data from Who into self
                        //existing values will be replaced by imported ones
                        //or they'll be kept unchanged
                        //min, max & init values of tables' columns will remain
                        //unchanged
                        //no data conversion will be done
                        //cannot assimilate self

                function TablesCount:Integer; virtual; abstract;
                function AddTable(TblName:WideString):TXDBTable; dynamic; abstract;
                        //creates new table, if table already exists
                        //if desired then also creates columns
                        //then returns nil

                property Tables[index:integer]:TXDBTable read GetTable;
                property NamedTables[TblName:WideString]:TXDBTable
                        read GetNamedTable; default;
              end;

    TXDBTable = class
                protected
                  FName:WideString;
                  FTemp:pointer;
                  procedure SetName(NewName:WideString); virtual; abstract;
                  function GetColumn(index:integer):TXDBColumn; virtual; abstract;
                  function GetData(ACol, ARec:integer):TXDBDataObject; virtual; abstract;
                  function GetRecsCount:integer; virtual; abstract;
                public
                  function NamedCol(ColName:WideString):integer; virtual; abstract;
                        //returns position of column named ColName
                        //first column is at position 0
                        //if column doesn't exists returns negative value

                  function AddColumn(ColName:WideString):TXDBColumn; dynamic; abstract;
                        //creates new column in table
                        //this values will be initialized
                        // Name: ColName
                        // Kind: RawData
                        //rest of values will be zeros or empty strings
                        //returns newly created column or nil(failure)
                  procedure MoveColumn(FromPos, ToPos:integer); dynamic; abstract;
                        //moves column from position FromPos to position ToPos
                        //also moves every column between FromPos& ToPos
                        //first column is at position 0
                  function ColumnsCount:integer; virtual; abstract;

                  procedure CreateRecord(Index:integer); virtual; abstract;
                        //creates new empty record at position index
                        //if index = -1 then record is appended after
                        //last positioned record
                        //position of first record is 0

                  procedure DeleteRecord(index:integer); virtual; abstract;
                        //deletes record at position index
                  procedure MoveRecord(FromPos, ToPos:integer); virtual; abstract;
                        //moves record from position FromPos to position ToPos
                        //also moves every record between FromPos& ToPos
                  procedure XCHGRecord(FromPos, ToPos:integer); virtual; abstract;
                        //exchanges records at positions FromPos & ToPos

                  property Name:WideString read FName write SetName;
                  property Temp:pointer read FTemp write FTemp;
                        //possibility to associate some data with table
                        //when table is created, Temp is nil
                  property Columns[index:integer]:TXDBColumn read GetColumn;

                  property Data[ColNo, RecNo:integer]:TXDBDataObject
                        read GetData; default;
                        //access to data in table's column positioned at ColNo
                        //in record at position RecNo
                        //!!!!! THIS OBJECT CAN NOT BE FREE FROM MEMORY !!!!!!
                        //if invalid data referenced then returns nil
                        //reault of this code
                        //var pom:TXDBDataObject;
                        //  pom:=Data[0,0];
                        //  Data[1,4].AsDouble:=Data[2,3].AsDouble+pom.AsDouble;
                        //is not equal to the following code
                        //var pom:double;
                        //  pom:=Data[0,0].AsDouble;
                        //  Data[1,4].AsDouble:=Data[2,3].AsDouble+pom;
                        //but the first code is equal to this one
                        //Data[1,4].AsDouble:=2*Data[2,3].AsDouble;
                  property Records:integer read GetRecsCount; 
                        //count of records in table
                end;


    TXDBColumn = class
                 protected
                   FName:WideString;
                   FComment,
                   FInMask,
                   FOutMask:WideString;
                   FMinLength,
                   FMaxLength:Int64;
                   procedure SetKind(NewKind:Int64); virtual; abstract;
                   function GetMinVal:TXDBDataObject; virtual; abstract;
                   function GetMaxVal:TXDBDataObject; virtual; abstract;
                   function GetDefVal:TXDBDataObject; virtual; abstract;
                   function GetKind:Int64; virtual; abstract;
                   function GetInMask:WideString; virtual; abstract;
                   function GetOutMask:WideString; virtual; abstract;                   
                 public
                   procedure Assign(NewDesign:TColumnDesignTX); overload; virtual; abstract;
                   procedure Assign(ACol:TXDBColumn); overload; virtual; abstract;                    
                        //changes properties' values to new ones

                   property Name:WideString read FName write FName;
                   property Kind:Int64 read GetKind write SetKind;
                        //unknown data type is changed to RawData
                   property Comment:WideString read FComment write FComment;
                   property InputMask:WideString read GetInMask write FInMask;
                   property OutputMask:WideString read GetOutMask write FOutMask;
                   property MinLength:Int64 read FMinLength write FMinLength;
                   property MaxLength:Int64 read FMaxLength write FMaxLength;
                   property MinValue:TXDBDataObject read GetMinVal;
                   property MaxValue:TXDBDataObject read GetMaxVal;
                   property InitialValue:TXDBDataObject read GetDefVal;
                 end;


    TXDBDataObject = class //representation of data object as part of one record
                           //and its conversions
                     protected
                       FRaw:Pointer;
                       function GetRaw(BSz:integer):Pointer; virtual; abstract;
                       procedure SetRaw(BSz:integer; ARaw:Pointer); virtual; abstract;
                       function GetInt64:Int64; virtual; abstract;
                       procedure SetInt64(AInt64:Int64); virtual; abstract;
                       function GetDouble:Double; virtual; abstract;
                       procedure SetDouble(ADouble:Double); virtual; abstract;
                       function GetBoolean:Boolean; virtual; abstract;
                       procedure SetBoolean(ABoolean:Boolean); virtual; abstract;
                       function GetCurrency:Currency; virtual; abstract;
                       procedure SetCurrency(ACurrency:Currency); virtual; abstract;
                       function GetDateTime:TDateTime; virtual; abstract;
                       procedure SetDateTime(ADateTime:TDateTime); virtual; abstract;
                       function GetWide:WideString; virtual; abstract;
                       procedure SetWide(AWide:WideString); virtual; abstract;
                     public
                       function DataSize:integer; virtual; abstract;
                                //count of bytes allocated at AsBuffer

                       procedure Assign(AData:TXDBDataObject); virtual; abstract;
                                //copies data buffer from AData as xdbdtRawData
                                //=> keeps non-standard data lengths !
                                //AData = nil sets our value to NULL
                                //can not assign self

                       property PureRawData:Pointer read FRaw;
                                //points to internal data buffer when the data
                                //are hold, can be invalid value => verify it
                                //by calling DataSize

                       //followings are properties for data coneversions
                       //when read and/or write uses Copy On Write technique
                       property AsBuffer[BufSize:Integer]:Pointer read GetRaw write SetRaw;
                                //when AsBuffer[]:= BufSize determines the count
                                //of bytes that will be written
                                //if Kind is for example Integer then BufSize
                                //must at least 8 and 9th and more bytes are
                                //ignored;
                                //if BufSize = 0 then record will become empty
                                //if AsBuffer[BufSize]:=nil then BufSize
                                //if size of memory to be allocated for data
                                //content of memory can vary with every alloc

                                //when:=AsBuffer[] BufSize determines count of
                                //bytes to be read; if BufSize = 0 then
                                //all bytes will be read
                       property AsInt64:Int64 read GetInt64 write SetInt64;
                       property AsDouble:Double read GetDouble write SetDouble;
                       property AsBoolean:Boolean read GetBoolean write SetBoolean;
                       property AsCurrency:Currency read GetCurrency write SetCurrency;
                       property AsDateTime:TDateTime read GetDateTime write SetDateTime;
                       property AsWideString:WideString read GetWide write SetWide;
                     end;

    TFilter = class
              public
                function IOCtl(Code:Integer; Data:pointer):integer; virtual; abstract;
                        //executes with specific operation
                        //Code - number of operation
                        //Data - points to data needed to operation
                        //       when Data = nil only tests if Code is supported
                        //returns zero if operation succesfull or operation
                        //is supported else returns non-zero value
                        //should be used to implement checking of database
                        //if database is supported by this filter and so on
              end;

implementation

end.
