unit editwnd;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  UserLevel, ALinks, ExtCtrls, ComCtrls, Menus, ActnList, ImgList;

type
  TDBWnd = class(TADBWnd)
    DBExplorer: TTreeView;
    DBExPopup: TPopupMenu;
    ActionList1: TActionList;
    AddTable: TAction;
    AddGen: TAction;
    RemoveGen: TAction;
    Pidat1: TMenuItem;
    Zruit1: TMenuItem;
    RenameGen: TAction;
    DelTable: TAction;
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure DBExplorerEditing(Sender: TObject; Node: TTreeNode;
      var AllowEdit: Boolean);
    procedure AddGenExecute(Sender: TObject);
    procedure AddTableExecute(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure DBExplorerEdited(Sender: TObject; Node: TTreeNode;
      var S: String);
    procedure DBExplorerDblClick(Sender: TObject);
    procedure DelTableExecute(Sender: TObject);
    procedure RemoveGenExecute(Sender: TObject);
    procedure DBExplorerKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
  private
    FMain:TAMain;       //window which has attached us
    CanONWMSize:Boolean; //semaphor for WM_SIZE
    procedure WMSize(var msg:TWMSize); message WM_SIZE;
                                //to get main window not overlapped on maximaze
  protected
    FDBase:TXDBase;     //our db we work with
    TablesNode:TTreeNode; //node representing tables
    procedure CreateNewTable;   //creates new empty and named table
    procedure DeleteTable(Node: TTreeNode);
                        //deletes table which is represented by Node
    procedure AriseTableWnd(Node: TTreeNode); 
                        //invokes table edit window;
                        //the table is represented by Node
  public
    constructor Create(AOwner:TComponent; DBase:TXDBase); reintroduce;
    destructor Destroy; override;

    procedure Attach(At:TAMain); override;
    procedure FlushDB; override; //writes the data to the disk
  end;

implementation

uses main, strs, mconst, XDBRetCodes, tbledit;

{$R *.DFM}

constructor TDBWnd.Create(AOwner:TComponent; DBase:TXDBase);
var i:integer;
begin
  inherited Create(AOwner);


  FDBase:=DBase;

  CanONWMSize:=true;

  TablesNode:=DBExplorer.Items.Add(nil, TablesStr);
  for i:=0 to FDBase.TablesCount-1 do
    DBExplorer.Items.AddChild(TablesNode, FDBase.Tables[i].Name);

  left:=Screen.Width-Width;
  if MainWnd.WindowState <> wsMaximized then Top:=0
     else Top:=MainWnd.Height;
end;

destructor TDBWnd.Destroy;
var i:integer;
    pom:TForm;
begin
  for i:=0 to FDBase.TablesCount-1 do
     begin
       pom:=TForm(FDBase.Tables[i].Temp);
       if pom<>nil then pom.Release;
     end;

  FDBase.Free;

  inherited;
end;

procedure TDBWnd.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  FMain.Dettach(Self);
  Action:=caFree;
end;

procedure TDBWnd.Attach(At:TAMain);
begin
  FMain:=At;
end;

procedure TDBWnd.WMSize(var msg:TWMSize);
begin
  //to keep main window not overlapped, but only if it is maximized
  if (msg.SizeType = SIZE_MAXIMIZED) and (MainWnd.WindowState = wsMaximized)
     and CanONWMSize then
     begin
       CanONWMSize:=false;
       inherited;
       Top:=MainWnd.Top+MainWnd.Height;
       Height:=Height-Top;
       CanONWMSize:=true;
     end else
         inherited;
end;


procedure TDBWnd.DBExplorerEditing(Sender: TObject; Node: TTreeNode;
  var AllowEdit: Boolean);
begin
  if Node = TablesNode then AllowEdit:=false;
end;

procedure TDBWnd.AddGenExecute(Sender: TObject);
begin
  AddTable.Execute;
end;

procedure TDBWnd.AddTableExecute(Sender: TObject);
begin
  CreateNewTable;
end;

procedure TDBWnd.CreateNewTable;
var NName:String;

begin
  NName:=NewTableStr;
  if not InputQuery(TblCreationStr, InputNameStr, NName) then exit;

  //table doesn't exists yet=>we can create ot
  if FDBase.AddTable(NName) = nil then
     begin   //a table couldn't be created
       case FMain.MainEngine.LastError of
         xdbrcOK:;
         xdbrcTableAlreadyExists:Application.MessageBox(TblAlrdExistsStr, InfoStr, InfoOK);
         xdbrcUnableToCreateTable:Application.MessageBox(CannotCreateTblStr, ErrorStr, ErrorOK);
       else
          Application.MessageBox(UnexpectedErStr, WarningStr, WarningOK);
       end;

       exit;
     end;

  //a table is created=>add it to the list
  DBExplorer.Items.AddChild(TablesNode, NName);
end;

procedure TDBWnd.FormActivate(Sender: TObject);
begin
  //register ourselves as an active database
  if FMain<>nil then FMain.ActiveDB:=self;
end;

procedure TDBWnd.FlushDB;
begin
  try
    FDBase.Flush;
  except
{    case MainWnd.MainEngine.LastError of
    else}
       Application.MessageBox(UnexpectedErStr, ErrorStr, ErrorOK);
{    end;}
  end;
end;

procedure TDBWnd.DBExplorerEdited(Sender: TObject; Node: TTreeNode;
  var S: String);
var pom:TXDBTable;
begin
  if Node.Parent = TablesNode then
     begin  //rename a table
       //get its name
       pom:=FDBase[Node.Text];

       //a set the new text
       try
         pom.Name:=S;
       except
         case FMain.MainEngine.LastError of
            xdbrcCannotRenameTable:Application.MessageBox(CannotRnTableStr, WarningStr, WarningOK);
            xdbrcInvalidName:Application.MessageBox(InvalidNameEnteredStr, WarningStr, WarningOK);
            xdbrcTableAlreadyExists:Application.MessageBox(TblAlrdExistsStr, WarningStr, WarningOK);
         else
            Application.MessageBox(UnexpectedErStr, WarningStr, WarningOK);
         end;
         S:=Node.Text;
         exit;
       end;

       //if needed, set new caption of edit window
       if pom.Temp <> nil then TForm(pom.Temp).Caption:=caption +' ['+s+']';
     end;
end;

procedure TDBWnd.DBExplorerDblClick(Sender: TObject);
begin
  AriseTableWnd(DBExplorer.Selected);
end;

procedure TDBWnd.AriseTableWnd(Node: TTreeNode);
var clickedTbl:TXDBTable;
begin
  //verify selected now
  if Node.Parent <> TablesNode then exit;

  clickedTbl:=FDBase[Node.Text];
  if clickedTbl = nil then exit; //something has failed

  with ClickedTbl do
    if Temp = nil then
       begin
         Temp:=TTblEditWnd.Create(Self, clickedTbl);
         TForm(Temp).Caption := caption +' ['+node.text+']';
       end else
          if TForm(Temp).WindowState = wsMinimized then
             SendMessage(TForm(Temp).handle, WM_SYSCOMMAND, SC_RESTORE, 0)
             else TForm(Temp).Show;
end;


procedure TDBWnd.DeleteTable(Node: TTreeNode);
var tbl:TXDBTable;
begin
  if Node.Parent <> TablesNode then exit;
  if Application.MessageBox(PChar(DeleteTblQstStr+' "'+Node.Text+'"?'),
                            QuestionStr, QuestionYesNo) <> IDYes then exit;

  //get the table handle
  tbl:=FDBase[Node.Text];
  if tbl = nil then exit; //the table doesn't exists => nothing to deleet

  //if the window is opened the release it
  if tbl.Temp<>nil then TForm(tbl.Temp).Release;

  //and free the table (from the memory and the database)
  tbl.Free;

  //give the invalid noe away
  Node.Delete;
end;

procedure TDBWnd.DelTableExecute(Sender: TObject);
begin
  DeleteTable(DBExplorer.Selected);
end;

procedure TDBWnd.RemoveGenExecute(Sender: TObject);
begin
  DelTableExecute(Sender);
end;

procedure TDBWnd.DBExplorerKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  //catch the events
  case Key of
      vk_Delete:DeleteTable(DBExplorer.Selected); //table delete
      vk_Return:AriseTableWnd(DBExplorer.Selected);//show table's window
  end;
end;

end.
