unit main;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  UserLevel, Menus, ToolWin, ComCtrls, MenuBar, ExtCtrls, ImgList, ActnList,
  editwnd, UserConsts, ALinks;

type
  TMainWnd = class(TAMain)
    MainMenu: TMainMenu;
    Soubor1: TMenuItem;
    Nov1: TMenuItem;
    ImageList: TImageList;
    ActionList: TActionList;
    NewDB: TAction;
    CreateDialog: TOpenDialog;
    AboutMnu: TMenuItem;
    Okna1: TMenuItem;
    N1: TMenuItem;
    Konec1: TMenuItem;
    OpenDB: TAction;
    Otevtdatabzi1: TMenuItem;
    FlushDB: TAction;
    Zapsatdatabzi1: TMenuItem;
    OpenDialog: TOpenDialog;
    ToolBar1: TToolBar;
    ToolButton1: TToolButton;
    ToolButton2: TToolButton;
    ToolButton3: TToolButton;
    procedure NewDBExecute(Sender: TObject);
    procedure Konec1Click(Sender: TObject);
    procedure OpenDBExecute(Sender: TObject);
    procedure FlushDBExecute(Sender: TObject);
  private
    procedure FChildMenu(Sender:TObject);
        //database menu click handler
  protected
     FImportFilters,            //some filters doesn't need to support saving
     FExportFilters:array of integer;   //or reading;
                                  //CreateDialog.Filters = 2, menaser
                                //filter MainEngine.Filters[ImportFilters[2]]
     FChildren:array of TADBWnd;//previously created windows; indexes must
                                //match to order of items in database menu
     procedure CreateNewDBase;  //creates mew empty database
     procedure OpenDBase;       //loads existing database
     procedure SaveDBase;       //flushes active database
     procedure CreateChild(DB:TXDBase; Capt:String);
                ///creates slave window and adds it into the reference list
  public
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;

    procedure Dettach(Who:TADBWnd); override; //removes links to who
  end;

var
  MainWnd: TMainWnd;

implementation

uses strs, mconst, xdbRetCodes;

//we have to import only one function
function CreateXDBEngine(CacheSize:integer):TXDBEngine;    external 'XDBEngine.DLL';

{$R *.DFM}

constructor TMainWnd.Create(AOwner:TComponent);
var i, j, k:integer;
    s:string;
begin
  inherited;

  //db interface initialization
  FMainEngine:=CreateXDBEngine(4*1024*1024); //4 MB big cache

  //setup output filters
  k:=0;
  s:='';
  j:=FMainEngine.FiltersCount;
  SetLength(FExportFilters, j);
  for i:=0 to j-1 do
     with FMainEngine.Filters[i] do
        if CanExport and (Flags and fsNaming <> 0) then
           begin
             FExportFilters[k]:=i;
             inc(k);
             s:=s+ExportName+'|'+ExportMask+'|';
           end;

  SetLength(FExportFilters, k);
  CreateDialog.Filter:=copy(s, 1, Length(s)-1); //kill  '|' at the end

  //setup input filters
  k:=0;
  s:='';
  j:=FMainEngine.FiltersCount;
  SetLength(FImportFilters, j);
  for i:=0 to j-1 do
     with FMainEngine.Filters[i] do
        if CanImport and (Flags and fsNaming <> 0) then
           begin
             FImportFilters[k]:=i;
             inc(k);
             s:=s+ImportName+'|'+ImportMask+'|';
           end;

  SetLength(FImportFilters, k);
  OpenDialog.Filter:=copy(s, 1, Length(s)-1);  //kill  '|' at the end

  FChildren:=nil;
  FActiveDB:=nil;
end;

destructor TMainWnd.Destroy;
begin
  FMainEngine.Free;

  inherited;
end;

procedure TMainWnd.FChildMenu(Sender:TObject);
var pom:TForm;
begin     //redraw or restore window
  pom:=FChildren[(Sender as TMenuItem).MenuIndex];
  if pom.WindowState = wsMinimized then
     SendMessage(pom.handle,WM_SYSCOMMAND, SC_RESTORE, 0)
     else pom.Show;
end;

procedure TMainWnd.CreateChild(DB:TXDBase; Capt:String);
var Child:TADBWnd;
    i:integer;
    item:TMenuItem;
begin
  Child:=TDBWnd.Create(Application, DB);
  Child.Caption:=Capt;
  Child.Attach(Self);
  FActiveDB:=Child;

  //add window to the list
  i:=Length(FChildren);
  SetLength(FChildren, i+1);
  FChildren[i]:=Child;

  Item:=TMenuItem.Create(Self);
  Item.Caption:=Capt;
  Item.OnClick:=FChildMenu;
  WindowMenu.Add(Item);
end;


procedure TMainWnd.Dettach(Who:TADBWnd);
var i,j:integer;
begin
  //find inxes of window to remove
  for i:=0 to High(FChildren) do
      if FChildren[i] = Who then break;

  //i determines number to remove
  WindowMenu.Items[i].Free;
  for j:=i to High(FChildren)-1 do
    FChildren[j]:=FChildren[j+1];
  SetLength(FChildren, High(FChildren));

  //we could remove active database
  if Who = FActiveDB then FActiveDB:=nil;
end;


procedure TMainWnd.CreateNewDBase;
var DBase:TXDBase;
    {tmp:TOpenWizardRec;
    wizarded:boolean;}
begin
  //we have to get filename & index
  if not CreateDialog.Execute then exit;

{  //try to run wizard
  tmp.Engine:=FMainEngine;
  tmp.DB:=nil;
  wizarded:=false;
  try
    if FMainEngine.FilterIOCtl(CreateDialog.FilterIndex-1, 0, @tmp) = 0 then
      wizarded:=true;
  finally
    if wizarded then DBase:=tmp.DB;
  end;
 }

  //and then we are able to create database
  if {not(wizarded) and} (FMainEngine.OpenDBase(CreateDialog.FileName, true,
         FExportFilters[CreateDialog.FilterIndex-1], DBase) <> xdbrcOK) then
     begin
       Application.MessageBox(CannotCreateDBStr, WarningStr, WarningOK);
       exit;
     end;

  {if DBase<>nil then }CreateChild(DBase, CreateDialog.FileName);
end;

procedure TMainWnd.NewDBExecute(Sender: TObject);
begin
  CreateNewDBase;
end;

procedure TMainWnd.Konec1Click(Sender: TObject);
begin
  close;
end;

procedure TMainWnd.OpenDBExecute(Sender: TObject);
begin
  OpenDBase;
end;

procedure TMainWnd.SaveDBase;
begin
  if FActiveDB = nil then exit; //no one to be flushed
  FActiveDB.FlushDB;
end;

procedure TMainWnd.FlushDBExecute(Sender: TObject);
begin
  SaveDBase;
end;

procedure TMainWnd.OpenDBase;
var DBase:TXDBase;
begin
  //get filename
  if not OpenDialog.Execute then exit;

  //open database
  if FMainEngine.OpenDBase(OpenDialog.FileName, false,
         FImportFilters[OpenDialog.FilterIndex-1], DBase) <> xdbrcOK then
     begin
       Application.MessageBox(CannotOpenDBStr, WarningStr, WarningOK);
       exit;
     end;

  CreateChild(DBase, OpenDialog.FileName);
end;


end.
