unit tbledit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ALinks, UserLevel, ComCtrls, ExtCtrls, Grids, StdCtrls, Spin, Buttons,
  ImgList, ActnList, Mask, Menus;

type
  TTblEditWnd = class(TDBCommonEditWnd)
    PageControl: TPageControl;
    EditSheet: TTabSheet;
    DesignSheet: TTabSheet;
    Splitter1: TSplitter;
    DesignGroup: TGroupBox;
    Label1: TLabel;
    ColsView: TListView;
    DataKind: TComboBox;
    Label2: TLabel;
    Name: TEdit;
    Label3: TLabel;
    Label4: TLabel;
    Label5: TLabel;
    Label6: TLabel;
    Label7: TLabel;
    Label8: TLabel;
    InMask: TEdit;
    OutMask: TEdit;
    MinLng: TSpinEdit;
    MaxLng: TSpinEdit;
    MinVal: TMaskEdit;
    MaxVal: TMaskEdit;
    Label9: TLabel;
    Note: TEdit;
    AddSBtn: TSpeedButton;
    DelColBtn: TSpeedButton;
    AplSBtn: TSpeedButton;
    Label10: TLabel;
    Label11: TLabel;
    DefVal: TMaskEdit;
    ActionList1: TActionList;
    AddCol: TAction;
    ImageList1: TImageList;
    ApplyDsg: TAction;
    DelCols: TAction;
    CUpBtn: TSpeedButton;
    CDwnBtn: TSpeedButton;
    ColUp: TAction;
    ColDown: TAction;
    DGrid: TDrawGrid;
    DataGridPM: TPopupMenu;
    InsertRec: TAction;
    AppendRec: TAction;
    DelRec: TAction;
    Pidatdek1: TMenuItem;
    InsertRec1: TMenuItem;
    DelRec1: TMenuItem;
    N1: TMenuItem;
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure ColsViewResize(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure AddColExecute(Sender: TObject);
    procedure ApplyDsgExecute(Sender: TObject);
    procedure ColsViewSelectItem(Sender: TObject; Item: TListItem;
      Selected: Boolean);
    procedure ColsViewDblClick(Sender: TObject);
    procedure DelColsExecute(Sender: TObject);
    procedure InMaskChange(Sender: TObject);
    procedure MinValChange(Sender: TObject);
    procedure MaxValChange(Sender: TObject);
    procedure DefValChange(Sender: TObject);
    procedure ColUpExecute(Sender: TObject);
    procedure ColDownExecute(Sender: TObject);
    procedure DGridDrawCell(Sender: TObject; ACol, ARow: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure DGridGetEditMask(Sender: TObject; ACol, ARow: Integer;
      var Value: String);
    procedure DGridGetEditText(Sender: TObject; ACol, ARow: Integer;
      var Value: String);
    procedure DGridSetEditText(Sender: TObject; ACol, ARow: Integer;
      const Value: String);
    procedure AppendRecExecute(Sender: TObject);
    procedure InsertRecExecute(Sender: TObject);
    procedure DelRecExecute(Sender: TObject);
  private
    FMnVDsgd,        //true if changed value: minimum
    FMxVDsgd,                                        //maximum
    FDfVDsgd:Boolean;                                   //init
  protected
    FTable:TXDBTable;
    LastDesigned:TXDBColumn; //recently designed column
    procedure GoDesignDisabled(ForceEnable:Boolean = false);
              //disables editation of the table's column when it is being designed
              //it sets ForceEnable enabled at controls in DesignGroup
    procedure GoDesignCol(ColNo:integer);
                //sets the column in order ColNo in the table edit controls
    procedure CreateNewCol;
                  //creates new column for a table and prepares it for the editation
    procedure ApplyDesign;
                //applies desired design to LastDesigned column
    procedure RemoveCols; //kills all selected columns
    procedure LoadCols;  //loads names of columns into the ColsView
  public
    constructor Create(AOwner:TComponent; ATable:TXDBTable); reintroduce;
    procedure GoEdit; override;
    procedure GoDesign; override;
  end;

implementation

uses Strs, UserStructs, main, xdbRetCodes, mconst;

{$R *.DFM}

constructor TTblEditWnd.Create(AOwner:TComponent; ATable:TXDBTable);
begin
  inherited Create(AOwner);

  FTable:=ATable;
  LastDesigned:=nil;
//  Caption:=TableStr + ' - ['+FTable.Name+']';

  LoadCols;
end;

procedure TTblEditWnd.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  Action:=caFree;
  FTable.Temp:=nil;
end;

procedure TTblEditWnd.GoEdit;
begin
  PageControl.ActivePage:=EditSheet;
end;

procedure TTblEditWnd.GoDesign;
begin
  PageControl.ActivePage:=DesignSheet;
end;

procedure TTblEditWnd.ColsViewResize(Sender: TObject);
begin
  with ColsView do
    Columns[2].Width:=ClientWidth-Columns[1].Width-Columns[0].Width;
end;

procedure TTblEditWnd.GoDesignDisabled(ForceEnable:Boolean = false);
begin
  Name.Enabled:=ForceEnable;
  InMask.Enabled:=ForceEnable;
  OutMask.Enabled:=ForceEnable;
  MinVal.Enabled:=ForceEnable;
  MaxVal.Enabled:=ForceEnable;
  DefVal.Enabled:=ForceEnable;
  Note.Enabled:=ForceEnable;
  DataKind.Enabled:=ForceEnable;
  MinLng.Enabled:=ForceEnable;
  MaxLng.Enabled:=ForceEnable;

  AddSBtn.Enabled:=ForceEnable or (ColsView.Items.Count = 0);
  CUpBtn.Enabled:=ForceEnable or (ColsView.Items.Count = 0);
  CDwnBtn.Enabled:=ForceEnable or (ColsView.Items.Count = 0);
  AplSBtn.Enabled:=ForceEnable;
  DelColBtn.Enabled:=ForceEnable and (ColsView.Items.Count <> 0);

  if not ForceEnable then LastDesigned:=nil;

  DesignGroup.Caption:=ColDisabledStr;
end;

procedure TTblEditWnd.GoDesignCol(ColNo:integer);
var pom:TXDBColumn;
begin
  //in first, it is needed to get entered colimnt
  pom:=FTable.Columns[ColNo];
  if pom = nil then exit; //invalid column

  Name.Text:=pom.Name;
  InMask.Text:=pom.InputMask;
  OutMask.Text:=pom.OutputMask;
  Note.Text:=pom.Comment;
  MinLng.Value:=pom.MinLength;
  MaxLng.Value:=pom.MaxLength;

  MinVal.Text:=pom.MinValue.AsWideString;
  MaxVal.Text:=pom.MaxValue.AsWideString;
  DefVal.Text:=pom.InitialValue.AsWideString;

  FMnVDsgd:=False;
  FMxVDsgd:=False;
  FDfVDsgd:=False;

  DataKind.ItemIndex:=pom.Kind-1;

  with ColsView.Items[ColNo] do
    begin
      Caption:=Name.Text;
      SubItems[0]:=DataKind.Text;
      SubItems[1]:=Note.Text;
    end;

  GoDesignDisabled(true);
  DesignGroup.Caption:=pom.Name;

  LastDesigned:=pom;
end;

procedure TTblEditWnd.CreateNewCol;
var pom:TXDBColumn;
    li:TListItem;
    s:string;
    i:integer;

begin
  try
    s:=ColumnStr+' ['+FormatDateTime('m.d.yyyy, hh:mm.zzz', Now)+']';
    pom:=FTable.AddColumn(s);
    pom.Kind:=xdbdtText;
    LastDesigned:=pom;
    with ColsView do
      if (Selected = nil) or (selected = Items[Items.Count-1]) then
        li:=ColsView.Items.Add
        else begin
               i:=Items.IndexOf(Selected);
               li:=Items.Insert(i);
               FTable.MoveColumn(FTable.ColumnsCount-1, i);
             end;
    li.Caption:=s;
    li.SubItems.Add('');
    li.SubItems.Add('');    
    GoDesignCol(FTable.ColumnsCount-1);
    LoadCols;
  except
    case MainWnd.MainEngine.LastError of
      xdbrcColAlreadyExists:Application.MessageBox(ColAlrdExistsStr, ErrorStr, ErrorOK);
    else
      Application.MessageBox(UnexpectedErStr, ErrorStr, ErrorOK);
    end;
  end;
end;

procedure TTblEditWnd.FormCreate(Sender: TObject);
begin
  DataKind.ItemIndex:=0;
  AddCol.ShortCut:=ShortCut(vk_Add, [ssCtrl]);
  DelCols.ShortCut:=ShortCut(vk_Subtract, [ssCtrl]);
  ColUp.ShortCut:=ShortCut(vk_Up, [ssCtrl]);
  ColDown.ShortCut:=ShortCut(vk_Down, [ssCtrl]);
  InsertRec.ShortCut:=ShortCut(vk_Insert, [ssCtrl, ssShift]);
  AppendRec.ShortCut:=ShortCut(ord('A'), [ssCtrl, ssShift]);
  DelRec.ShortCut:=ShortCut(vk_Delete, [ssCtrl, ssShift]);

  DGrid.RowCount:=FTable.Records+1;
end;

procedure TTblEditWnd.AddColExecute(Sender: TObject);
begin
  CreateNewCol;
end;

procedure TTblEditWnd.ApplyDesign;
var dsg:TColumnDesignTX;
    oldMn,
    oldMx,
    oldDf:Pointer; //old data values for the case of canceled editaion
    oMnS, oMxS, ODfS:integer; //their size in bytes

begin
  if LastDesigned = nil then exit;

  //prepare column's design and try to set it
  dsg.Name:=Name.Text;
  with dsg do
    begin
      Comment:=Note.Text;
      Kind:=DataKind.ItemIndex+1;
      InputMask:=InMask.Text;
      OutputMask:=OutMask.Text;
      MinLength:=MinLng.Value;
      MaxLength:=MaxLng.Value;
      MinValue:=MinVal.Text;
      MaxValue:=MaxVal.Text;
      InitialValue:=DefVal.Text;
    end;

  with LastDesigned do
    try
      oMnS:=MinValue.DataSize;
      oldMn:=MinValue.AsBuffer[0];
      oMxS:=MaxValue.DataSize;
      oldMx:=MaxValue.AsBuffer[0];
      oDfS:=InitialValue.DataSize;
      oldDf:=InitialValue.AsBuffer[0];


      if not FMnVDsgd then dsg.MinValue:='';
      if not FMxVDsgd then dsg.MaxValue:='';
      if not FDfVDsgd then dsg.InitialValue:='';

      Assign(dsg);

      if not FMnVDsgd then LastDesigned.MinValue.AsBuffer[oMnS]:=oldMn;
      if not FMxVDsgd then LastDesigned.MaxValue.AsBuffer[oMxS]:=oldMx;
      if not FDfVDsgd then LastDesigned.InitialValue.AsBuffer[oDfS]:=oldDf;


      FreeMem(oldMn);
      FreeMem(oldMx);
      FreeMem(oldDf);

    except
      case MainWnd.MainEngine.LastError of
        xdbrcColAlreadyExists:Application.MessageBox(ColAlrdExistsStr, ErrorStr, ErrorOK);
        xdbrcInvalidDColDsgnData:Application.MessageBox(InvalidDsgDataStr, ErrorStr, ErrorOK);
      else
        Application.MessageBox(UnexpectedErStr, ErrorStr, ErrorOK);
      end;
      exit;
    end;
  //load it to ensure ourselves
  with ColsView do
    GoDesignCol(Items.IndexOf(Selected));
end;

procedure TTblEditWnd.ApplyDsgExecute(Sender: TObject);
begin
  ApplyDesign;
end;

procedure TTblEditWnd.ColsViewSelectItem(Sender: TObject; Item: TListItem;
  Selected: Boolean);
begin
  GoDesignDisabled(ColsView.SelCount = 1);
end;

procedure TTblEditWnd.ColsViewDblClick(Sender: TObject);
begin
  with ColsView do
    GoDesignCol(Items.IndexOf(Selected));
end;

procedure TTblEditWnd.RemoveCols;
begin
  if ColsView.Selected = nil then
     begin  //don't know what to remove
       Application.MessageBox(NoColsSelectedStr, InfoStr, InfoOK);
       exit;
     end;

  //verify
  if Application.MessageBox(RemoveColsStr, QuestionStr, QuestionYesNo) <> IDYes
     then exit;

  //and remove
  with ColsView do
     while Selected<>nil do
        begin
          FTable.Columns[Items.IndexOf(Selected)].Free;
          Selected.Delete;
        end;

  GoDesignDisabled;
  LoadCols;
end;

procedure TTblEditWnd.DelColsExecute(Sender: TObject);
begin
  RemoveCols;
end;

procedure TTblEditWnd.InMaskChange(Sender: TObject);
begin
  MinVal.EditMask:=InMask.Text;
  MaxVal.EditMask:=InMask.Text;
  DefVal.EditMask:=InMask.Text;
end;

procedure TTblEditWnd.MinValChange(Sender: TObject);
begin
  FMnVDsgd:=true;
end;

procedure TTblEditWnd.MaxValChange(Sender: TObject);
begin
  FMxVDsgd:=true;
end;

procedure TTblEditWnd.DefValChange(Sender: TObject);
begin
  FDfVDsgd:=true;
end;

procedure TTblEditWnd.LoadCols;
var i:integer;
    li:TListItem;

begin
  ColsView.Items.Clear;
  //load the defintions of the columns
  for i:=0 to FTable.ColumnsCount-1 do
    begin
      li:=ColsView.Items.Add;
      li.Caption:=FTable.Columns[i].Name;
      li.SubItems.Add(DataKind.Items[FTable.Columns[i].Kind-1]);
      li.SubItems.Add(FTable.Columns[i].Comment);
    end;

  DGrid.ColCount:=FTable.ColumnsCount;
end;

procedure TTblEditWnd.ColUpExecute(Sender: TObject);
var i:integer;
begin
  if ColsView.Selected = nil then exit;

  //get the position of the selected column
  with ColsView do
    i:=Items.IndexOf(Selected)-1;
  //move it
  FTable.MoveColumn(i+1, i);
  //check it
  LoadCols;
  //select it once again
  if i<0 then i:=0;
  ColsView.Items[i].Selected:=true;
end;

procedure TTblEditWnd.ColDownExecute(Sender: TObject);
var i:integer;
begin
  if ColsView.Selected = nil then exit;

  //get the position of the selected column
  with ColsView do
    i:=Items.IndexOf(Selected)+1;
  //move it
  FTable.MoveColumn(i-1, i);
  //check it
  LoadCols;
  //select it once again
  if i>=ColsView.Items.Count then dec(i);
  ColsView.Items[i].Selected:=true;
end;

procedure TTblEditWnd.DGridDrawCell(Sender: TObject; ACol, ARow: Integer;
  Rect: TRect; State: TGridDrawState);
begin
  if ARow = 0 then      //the column's name is to be entered
     try
       if ACol<FTable.ColumnsCount then
         DGrid.Canvas.TextRect(Rect, Rect.Left+1, Rect.Top+1, FTable.Columns[ACol].Name);
     except
       DGrid.ColCount:=FTable.ColumnsCount;
//       refresh;
     end else
        begin  //we have to write the data
          dec(ARow);
          if ARow >= FTable.Records then exit; //we don't have so much data

          //draw
          DGrid.Canvas.TextRect(Rect, Rect.Left+1, Rect.Top+1, FTable[ACol, ARow].AsWideString);
        end;
end;

procedure TTblEditWnd.DGridGetEditMask(Sender: TObject; ACol,
  ARow: Integer; var Value: String);
begin
  try
    Value:=FTable.Columns[ACol].InputMask;
  except
  end;
end;

procedure TTblEditWnd.DGridGetEditText(Sender: TObject; ACol,
  ARow: Integer; var Value: String);
var pom:TXDBDataObject;
begin
  Dec(ARow);
  if ARow >= FTable.Records then exit; //there is not so much data

  pom:=FTable[ACol, ARow];
  if pom<>nil then Value:=pom.AsWideString
     else Value:='';
end;

procedure TTblEditWnd.DGridSetEditText(Sender: TObject; ACol,
  ARow: Integer; const Value: String);
begin
  try
    Dec(ARow);
    if ARow >= FTable.Records then exit; //there is not so much data

    FTable[ACol, ARow].AsWideString:=Value;
  except
//    Application.MessageBox(CannotEnterDataStr, WarningStr, WarningOK);
  end;
end;

procedure TTblEditWnd.AppendRecExecute(Sender: TObject);
begin
  FTable.CreateRecord(-1);
  DGrid.RowCount:=FTable.Records+1;
  if DGrid.RowCount > 1 then DGrid.FixedRows:=1;
  DGrid.Row:=DGrid.RowCount-1;
end;

procedure TTblEditWnd.InsertRecExecute(Sender: TObject);
begin
  FTable.CreateRecord(DGrid.Row-1);
  DGrid.RowCount:=FTable.Records+1;
  if DGrid.RowCount > 1 then DGrid.FixedRows:=1;
  DGrid.Refresh;
end;

procedure TTblEditWnd.DelRecExecute(Sender: TObject);
var i:integer;
begin
  if Application.MessageBox(DeleteRecQStr, WarningStr, QuestionYesNo) = IDNo
     then exit;

  i:=DGrid.Row-1;
  if i >= FTable.Records then exit; //invalid record id
  FTable.DeleteRecord(i);
  DGrid.RowCount:=DGrid.RowCount-1;
  DGrid.Refresh;
end;

end.
