//Part of XDBEngine source code.
//Version 1.2
//Created by Tomas Koutny.
//for more info see http://www.rawos.cz/xdbe/index.html

//defitions of objects that will be used at interface level of XDBCore's
//components when devoloping new main engine or database driver
unit AbstractLayer;

interface

uses UserLevel, windows, UserStructs, stream64, EngineLevel;


type
    //internally used data types
    TPositions = packed array of Int64;


    TAXDBase        = class;
    TAXDBTable      = class;
    TAXDBColumn     = class;
    TAXDBDataObject = class;

    //message of table rename
    //returns true if rename can has to be done
    TXDBTableRename = function(Sender:TAXDBTable; NewName:WideString):Boolean of object;

    //message that table destructor is called
    TXDBTableDestroy = procedure(Sender:TAXDBTable) of object;

    //message that table's column is being destroed
    TXDBColumnDestroy = procedure(Sender:TAXDBColumn) of object;
    //message that table's column is about to be renamed
    //returns true, if it can be done
    TXDBColumnRename = function(Sender:TAXDBColumn; NewName:WideString):boolean of object;


    TAXDBase = class(TXDBase)
               protected
                 FOwner:TAXEngine;
                 FDestroyingSelf:boolean; //true when we are destroying
                 FileStream:TStream64;    //but we will create TFileStream64
               public
                 constructor Create(AOwner:TAXEngine; DBName:String;
                                    CreateIt:Boolean); reintroduce; dynamic; abstract;
                 function TableRenameHandler(Sender:TAXDBTable;
                                 NewName:WideString):Boolean; dynamic; abstract;
                        //handler for message of table rename
                        //funtion must be called by table => public declaration
                 procedure TableDestroyHandler(Sender:TAXDBTable); dynamic; abstract;
                        //handler for message of table destroy
                                    
                 function PopPoss(FromPos, Count:Int64):TPositions; virtual; abstract;
                        //loads from db file table of offsets, there is
                        //count offsets and the place in db file will be marked
                        //freeed and free since next db save
                 function PushPoss(Poss:TPositions):Int64; virtual; abstract;
                        //saves into the db file table of offsets without
                        //any info about their count
                        //returns offset where table was saved
                        //place in db file will marked as at PopPoss
                        //Poss will not be freed from memory
                 procedure LoadData(FromPos, ID:Int64; Data:TAXDBDataObject); virtual; abstract;
                        //loads from db file part of record; part belongs
                        //to columnd ID, number of record is determined
                        //by position stored in db table object
                 procedure StoreData(var AtPos:Int64; ID:Int64; Data:TAXDBDataObject); virtual; abstract;
                        //tries to save Data at position AtPos, in way when
                        //data will belong to column ID
                        //returns position where the data has been written
                 procedure FreeRecord(AtPos:Int64); virtual; abstract;
                        //frees record at position AtPos and marks place
                        //in db file as free
                 procedure FreeBlock(From, Size:Int64); virtual; abstract;
                        //marks place in db file at position From with size Size
                        //as free

                 property DestroyingSelf:boolean read FDestroyingSelf;
                 property FStream:TStream64 read FileStream;
                 property Owner:TAXEngine read FOwner;
               end;

    TAXDBTable = class(TXDBTable)
                 protected
                   FOwner:TAXDBase; //database in which we belong
                 public
                   constructor Create(AOwner:TAXDBase); reintroduce; dynamic; abstract;
                   procedure ColumnDestroyHandler(Sender:TAXDBColumn); dynamic; abstract;
                   //handler for message of table's column destroy
                   function ColumnRenameHandler(Sender:TAXDBColumn;
                             NewName:WideString):boolean; dynamic; abstract;
                   //handler for massge of column rename

                   property Owner:TAXDBase read FOwner;
                 end;

    TAXDBColumn = class(TXDBColumn)
                  protected
                    function GetOwner:TAXDBTable; virtual; abstract;
                  public
                    constructor Create(AOwner:TAXDBTable); reintroduce; dynamic; abstract;
                    function VariableDataLength:boolean; virtual; abstract;
                        //returns true when Kind doesn't sign data with
                        //hard length

                    property Owner:TAXDBTable read GetOwner;
                  end;

    TAXDBDataObject = class(TXDBDataObject)
                      protected
                        FParse:Boolean;
                        FOwner:TAXDBColumn;
                      public
                        constructor Create(AOwner:TAXDBColumn); reintroduce; dynamic; abstract;

                        property Parse:Boolean read FParse write FParse;
                           //if true, input values will be parsed
                           //as specfied in table's column
                           //uses the min & max values
                        property Owner:TAXDBColumn read FOwner;
                      end;

implementation

end.
