//Part of XDB Core source code.
//Version 1.0
//Created by Tomas Koutny.
//for more info see http://www.rawos.cz/xdbe/index.html


//object representing data of part of one record for one column
unit DataObj;

interface

uses AbstractLayer, stream64, UserLevel;


type
  TXDBDataObj = class;

   //message that we wrote date into Sender
   TOnDataObjWritten = procedure(Sender:TXDBDataObj) of object;
   //message that there's demand to read Sender's data, but Sender
   //didn't load 'em yet
   TOnDataObjDataDemand = procedure(Sender:TXDBDataObj) of object;

   TXDBDataObj = class(TAXDBDataObject)
                 protected
                   FOnWritten:TOnDataObjWritten;
                   FOnDataDemand:TOnDataObjDataDemand;
                   FAllocated:integer; //how much bytes are allocated ata FRaw
                   function GetRaw(BSz:integer):Pointer; override;
                   procedure SetRaw(BSz:integer; ARaw:Pointer); override;
                   function GetInt64:Int64; override;
                   procedure SetInt64(AInt64:Int64); override;
                   function GetDouble:Double; override;
                   procedure SetDouble(ADouble:Double); override;
                   function GetBoolean:Boolean; override;
                   procedure SetBoolean(ABoolean:Boolean); override;
                   function GetCurrency:Currency; override;
                   procedure SetCurrency(ACurrency:Currency); override;
                   function GetDateTime:TDateTime; override;
                   procedure SetDateTime(ADateTime:TDateTime); override;
                   function GetWide:WideString; override;
                   procedure SetWide(AWide:WideString); override;

                   procedure ParseData; //checks data for min and max values
                 public
                   constructor Create(AOwner:TAXDBColumn); override;
                   destructor Destroy; override;

                   function DataSize:integer; override;

                   procedure Assign(AData:TXDBDataObject); override;
                          //copies data buffer from AData as xdbdtRawDaata
                          //=> keeps non-standard data lengths !
                          //AData = nil sets our value to NULL                          

                   function FileDataSize:Int64;
                        //returns how much bytes are needed to store data
                        //in table column header in db file
                   procedure LoadFromStream(Stream:TStream64);
                        //loads data from stream in dependence of FOwner.Kind
                        //loads data from header of column
                   procedure StoreToStream(Stream:TStream64); 
                        //writes info into stream
                        //presume write into column header

                   property OnDataWritten:TOnDataObjWritten
                        read FOnWritten write FOnWritten;
                   property OnDataDemand:TOnDataObjDataDemand
                        read FOnDataDemand write FOnDataDemand;
                 end;

implementation

uses UserStructs, Windows, SysUtils, XDBRetCodes, routines;


constructor TXDBDataObj.Create(AOwner:TAXDBColumn);
begin
  if AOwner = nil then Abort;
  FOwner:=AOwner;
  FAllocated:=0;
  FRaw:=nil;
  FOnWritten:=nil;
end;

destructor TXDBDataObj.Destroy;
begin
  if FRaw<>nil then FreeMem(FRaw);

  inherited;
end;

function TXDBDataObj.FileDataSize:Int64;
begin
  result:=8;
  if FOwner.Kind = xdbdtBoolean then result:=1;
  if FOwner.VariableDataLength then result:=result+Int64(FAllocated);
end;

procedure TXDBDataObj.LoadFromStream(Stream:TStream64);
var ToLoad:Int64;
begin
  //as first we have to know how much bytes we will load
  ToLoad:=8;
  if FOwner.Kind = xdbdtBoolean then ToLoad:=1;
  if FOwner.VariableDataLength then Stream.Read(ToLoad, 8);

  if FRaw<>nil then FreeMem(FRaw);
  FRaw:=nil;
  if ToLoad = 0 then exit;
  GetMem(Fraw, ToLoad);
  Stream.Read(FRaw^, ToLoad);
  FAllocated:=ToLoad;
end;

procedure TXDBDataObj.StoreToStream(Stream:TStream64);
var pom, pom2:Int64;
begin
  //we cann't forget that always will be save at least one byte
  pom:=Int64(FAllocated);
  if FOwner.VariableDataLength then Stream.Write(pom, 8);
  if FAllocated>0 then Stream.Write(FRaw^, FAllocated)
     else if not FOwner.VariableDataLength then
        begin
          pom2:=0; //by default we will return false or zero
          pom:=8;
          if FOwner.Kind = xdbdtBoolean then pom:=1;
          Stream.Write(pom2, pom);
        end;
end;

function TXDBDataObj.GetRaw(BSz:integer):Pointer;
begin
  if Assigned(FOnDataDemand) then FOnDataDemand(Self);

  FOwner.Owner.Owner.Owner.SetLastError(xdbrcOK);
  result:=nil;
  if FRaw = nil then exit; //nothing to copy
  if (FAllocated < BSz) or (Bsz = 0) then Bsz:=FAllocated;
        //how much bytes to copy
  FOwner.Owner.Owner.Owner.SetLastError(xdbrcUnexpected);
  GetMem(result, Bsz);
  Move(FRaw^, result^, Bsz);
  FOwner.Owner.Owner.Owner.SetLastError(xdbrcOK);
end;

procedure TXDBDataObj.SetRaw(BSz:integer; ARaw:Pointer);
begin
  if FRaw<>nil then FreeMem(FRaw);
  FRaw:=nil;
  FAllocated:=Bsz;
  if (ARaw = nil) and (FAllocated>0) then GetMem(FRaw, FAllocated);
  if (Bsz = 0) or (ARaw = nil) then exit;


  FOwner.Owner.Owner.Owner.SetLastError(xdbrcInvalidData);

  FAllocated:=8;
  with FOwner do
    case Kind of
      xdbdtRawData,
      xdbdtText     :FAllocated:=Bsz;
      xdbdtInteger,
      xdbdtCurrency,
      xdbdtDateTime,
      xdbdtReal     :if Bsz<8 then Abort;
      xdbdtBoolean  :FAllocated:=1;
  end;
  FOwner.Owner.Owner.Owner.SetLastError(xdbrcNotEnoughMemory);

  GetMem(FRaw, FAllocated);
  Move(ARaw^, FRaw^, FAllocated);

  FOwner.Owner.Owner.Owner.SetLastError(xdbrcInvalidData);
  ParseData;

  FOwner.Owner.Owner.Owner.SetLastError(xdbrcOK);
  if Assigned(FOnWritten) then FOnWritten(Self);
end;

function TXDBDataObj.GetInt64:Int64;
begin
  if Assigned(FOnDataDemand) then FOnDataDemand(Self);

  result:=0;
  if FRaw = nil then
{     begin
       if  FOwner.InitialValue<>Self then result:=FOwner.InitialValue.AsInt64;}
       exit;
{     end;}

  FOwner.Owner.Owner.Owner.SetLastError(xdbrcInvalidData);
  case FOwner.Kind of
     xdbdtRawData,
     xdbdtInteger  :begin
                      if FAllocated < 8 then Abort;
                      result:=PInt64(FRaw)^;
                    end;
     xdbdtReal     :begin
                      if FAllocated < 8 then Abort;
                      result:=round(PDouble(FRaw)^);
                    end;
     xdbdtBoolean  :begin
                      result:=-1;
                      if PByte(FRaw)^ = 0 then result:=0;
                    end;
     xdbdtCurrency :begin
                      if FAllocated < 8 then Abort;
                      result:=round(PCurrency(FRaw)^);
                    end;
     xdbdtDateTime :begin
                      if FAllocated < 8 then Abort;
                      result:=round(PDateTime(FRaw)^);
                    end;
     xdbdtText     :result:=StrToInt64(KillSpaces(AsWideString));
  else
    Abort;
  end;
  FOwner.Owner.Owner.Owner.SetLastError(xdbrcOK);
end;

procedure TXDBDataObj.SetInt64(AInt64:Int64);
var pomw:WideString;
begin
  if FRaw<>nil then FreeMem(FRaw);

  FAllocated:=8;
  case FOwner.Kind of
    xdbdtRawData,
    xdbdtInteger  :begin
                     GetMem(FRaw, 8);
                     PInt64(FRaw)^:=AInt64;
                   end;
    xdbdtReal,
    xdbdtDateTime :begin
                     GetMem(FRaw, 8);
                     PDouble(FRaw)^:=AInt64;
                   end;
    xdbdtBoolean  :begin
                     GetMem(FRaw, 1);
                     FAllocated:=1;
                     PByte(FRaw)^:=255;
                     if AInt64 = 0 then PByte(FRaw)^:=0;
                   end;
    xdbdtCurrency :begin
                     GetMem(FRaw, 8);
                     PCurrency(FRaw)^:=AInt64;
                   end;
    xdbdtText     :begin
                     if FOwner.OutputMask = '' then pomw:=IntToStr(AInt64)
                        else pomw:=FormatFloat(FOwner.OutputMask, AInt64);
                     FAllocated:=Length(pomw) shl 1;
                     Move(pointer(pomw)^, FRaw^, FAllocated);
                     pomw:='';
                   end;
  end;

  ParseData;
  FOwner.Owner.Owner.Owner.SetLastError(xdbrcOK);
  if Assigned(FOnWritten) then FOnWritten(Self);
end;

function TXDBDataObj.GetDouble:Double;
begin
  if Assigned(FOnDataDemand) then FOnDataDemand(Self);

  result:=0.0;
  if FRaw = nil then
{     begin
       if  FOwner.InitialValue<>Self then result:=FOwner.InitialValue.AsDouble;}
       exit;
{     end;                }

  FOwner.Owner.Owner.Owner.SetLastError(xdbrcInvalidData);
  case FOwner.Kind of
     xdbdtRawData,
     xdbdtReal,
     xdbdtDateTime :begin
                      if FAllocated < 8 then Abort;
                      result:=PDouble(FRaw)^;
                    end;
     xdbdtInteger:begin
                      if FAllocated < 8 then Abort;
                      result:=PInt64(FRaw)^;
                    end;
     xdbdtBoolean  :begin
                      result:=-1;
                      if PByte(FRaw)^ = 0 then result:=0;
                    end;
     xdbdtCurrency :begin
                      if FAllocated < 8 then Abort;
                      result:=PCurrency(FRaw)^;
                    end;
     xdbdtText     :result:=StrToFloat(KillSpaces(AsWideString));
  else
    Abort;
  end;
  FOwner.Owner.Owner.Owner.SetLastError(xdbrcOK);
end;

procedure TXDBDataObj.SetDouble(ADouble:Double);
var pomw:WideString;
begin
  if FRaw<>nil then FreeMem(FRaw);

  FAllocated:=8;
  case FOwner.Kind of
    xdbdtRawData,
    xdbdtReal,
    xdbdtDateTime  :begin
                     GetMem(FRaw, 8);
                     PDouble(FRaw)^:=ADouble;
                   end;

    xdbdtInteger:begin
                     GetMem(FRaw, 8);
                     PInt64(FRaw)^:=round(ADouble);
                   end;
    xdbdtBoolean  :begin
                     GetMem(FRaw, 1);
                     FAllocated:=1;
                     PByte(FRaw)^:=255;
                     if ADouble = 0 then PByte(FRaw)^:=0;
                   end;
    xdbdtCurrency :begin
                     GetMem(FRaw, 8);
                     PCurrency(FRaw)^:=ADouble;
                   end;
    xdbdtText     :begin
                     if FOwner.OutputMask = '' then pomw:=FloatToStr(ADouble)
                        else pomw:=FormatFloat(FOwner.OutputMask, ADouble);
                     FAllocated:=Length(pomw) shl 1;
                     Move(pointer(pomw)^, FRaw^, FAllocated);
                     pomw:='';
                   end;
  end;

  ParseData;
  FOwner.Owner.Owner.Owner.SetLastError(xdbrcOK);
  if Assigned(FOnWritten) then FOnWritten(Self);
end;


function TXDBDataObj.GetBoolean:Boolean;
begin
  if Assigned(FOnDataDemand) then FOnDataDemand(Self);

  result:=false;
  if FRaw = nil then
{     begin
       if  FOwner.InitialValue<>Self then result:=FOwner.InitialValue.AsBoolean;}
       exit;
{     end;}

  FOwner.Owner.Owner.Owner.SetLastError(xdbrcInvalidData);
  case FOwner.Kind of
     xdbdtRawData,
     xdbdtBoolean  :result:=PByte(FRaw)^ <> 0;
     xdbdtReal,
     xdbdtDateTime :begin
                      if FAllocated < 8 then Abort;
                      result:=PDouble(FRaw)^ <> 0;
                    end;
     xdbdtInteger:begin
                      if FAllocated < 8 then Abort;
                      result:=PInt64(FRaw)^ <> 0;
                    end;
     xdbdtCurrency :begin
                      if FAllocated < 8 then Abort;
                      result:=PCurrency(FRaw)^ <> 0;
                    end;
     xdbdtText     :result:=Char(PByte(Fraw)^) <> UpperCase(FOwner.Owner.Owner.Owner.NoString)[1];
  else
    Abort;
  end;
  FOwner.Owner.Owner.Owner.SetLastError(xdbrcOK);
end;

procedure TXDBDataObj.SetBoolean(ABoolean:Boolean);
var pomw:WideString;
begin
  if FRaw<>nil then FreeMem(FRaw);

  FAllocated:=8;
  case FOwner.Kind of
    xdbdtRawData,
    xdbdtBoolean   :begin
                     GetMem(FRaw, 1);
                     FAllocated:=1;
                     PByte(FRaw)^:=0;
                     if ABoolean then PByte(FRaw)^:=255;
                    end;
    xdbdtReal,
    xdbdtDateTime  :begin
                     GetMem(FRaw, 8);
                     PDouble(FRaw)^:=0;
                     if ABoolean then
                        if FOwner.Kind = xdbdtDateTime then PDateTime(FRaw)^:=now
                           else PDouble(FRaw)^:=-1;
                   end;
    xdbdtInteger:begin
                     GetMem(FRaw, 8);
                     PInteger(FRaw)^:=0;
                     if ABoolean then PInteger(FRaw)^:=-1;
                   end;
    xdbdtCurrency :begin
                     GetMem(FRaw, 8);
                     PCurrency(FRaw)^:=0;
                     if ABoolean then PCurrency(FRaw)^:=-1;
                   end;
    xdbdtText     :begin
                     if ABoolean then pomw:=FOwner.Owner.Owner.Owner.YesString
                        else pomw:=FOwner.Owner.Owner.Owner.NoString;
                     FAllocated:=Length(pomw) shl 1;
                     Move(pointer(pomw)^, FRaw^, FAllocated);
                     pomw:='';
                   end;
  end;

  ParseData;
  FOwner.Owner.Owner.Owner.SetLastError(xdbrcOK);
  if Assigned(FOnWritten) then FOnWritten(Self);
end;

function TXDBDataObj.GetCurrency:Currency;
begin
  if Assigned(FOnDataDemand) then FOnDataDemand(Self);

  result:=0.0;
  if FRaw = nil then
{     begin
       if  FOwner.InitialValue<>Self then result:=FOwner.InitialValue.AsCurrency;}
       exit;
{     end;   }

  FOwner.Owner.Owner.Owner.SetLastError(xdbrcInvalidData);
  case FOwner.Kind of
     xdbdtRawData,
     xdbdtCurrency :begin
                      if FAllocated < 8 then Abort;
                      result:=PCurrency(FRaw)^;
                    end;
     xdbdtReal,
     xdbdtDateTime :begin
                      if FAllocated < 8 then Abort;
                      result:=PDouble(FRaw)^;
                    end;
     xdbdtInteger:begin
                      if FAllocated < 8 then Abort;
                      result:=PInt64(FRaw)^;
                    end;
     xdbdtBoolean  :begin
                      result:=-1;
                      if PByte(FRaw)^ = 0 then result:=0;
                    end;
     xdbdtText     :result:=StrToCurr(KillSpaces(AsWideString));
  else
    Abort;
  end;
  FOwner.Owner.Owner.Owner.SetLastError(xdbrcOK);
end;

procedure TXDBDataObj.SetCurrency(ACurrency:Currency);
var pomw:WideString;
begin
  if FRaw<>nil then FreeMem(FRaw);

  FAllocated:=8;
  case FOwner.Kind of
    xdbdtRawData,
    xdbdtCurrency :begin
                     GetMem(FRaw, 8);
                     PCurrency(FRaw)^:=ACurrency;
                   end;
    xdbdtReal,
    xdbdtDateTime  :begin
                     GetMem(FRaw, 8);
                     PDouble(FRaw)^:=ACurrency;
                   end;

    xdbdtInteger:begin
                     GetMem(FRaw, 8);
                     PInt64(FRaw)^:=round(ACurrency);
                   end;
    xdbdtBoolean  :begin
                     GetMem(FRaw, 1);
                     FAllocated:=1;
                     PByte(FRaw)^:=255;
                     if ACurrency = 0 then PByte(FRaw)^:=0;
                   end;
    xdbdtText     :begin
                     if FOwner.OutputMask = '' then pomw:=CurrToStr(ACurrency)
                        else pomw:=FormatCurr(FOwner.OutputMask, ACurrency);
                     FAllocated:=Length(pomw) shl 1;
                     Move(pointer(pomw)^, FRaw^, FAllocated);
                     pomw:='';
                   end;
  end;

  ParseData;
  FOwner.Owner.Owner.Owner.SetLastError(xdbrcOK);
  if Assigned(FOnWritten) then FOnWritten(Self);
end;

function TXDBDataObj.GetDateTime:TDateTime;
begin
  if Assigned(FOnDataDemand) then FOnDataDemand(Self);

  result:=0.0;
  if FRaw = nil then
{     begin
       if  FOwner.InitialValue<>Self then result:=FOwner.InitialValue.AsDateTime;}
       exit;
{     end;}

  FOwner.Owner.Owner.Owner.SetLastError(xdbrcInvalidData);
  case FOwner.Kind of
     xdbdtRawData,
     xdbdtReal,
     xdbdtDateTime :begin
                      if FAllocated < 8 then Abort;
                      result:=PDateTime(FRaw)^;
                    end;
     xdbdtInteger:begin
                      if FAllocated < 8 then Abort;
                      result:=PInt64(FRaw)^;
                    end;
     xdbdtBoolean  :begin
                      result:=now;
                      if PByte(FRaw)^ = 0 then result:=0;
                    end;
     xdbdtCurrency :begin
                      if FAllocated < 8 then Abort;
                      result:=PCurrency(FRaw)^;
                    end;
     xdbdtText     :result:=StrToDateTime(KillSpaces(AsWideString));
  else
    Abort;
  end;
  FOwner.Owner.Owner.Owner.SetLastError(xdbrcOK);
end;

procedure TXDBDataObj.SetDateTime(ADateTime:TDateTime);
var pomw:WideString;
begin
  if FRaw<>nil then FreeMem(FRaw);

  FAllocated:=8;
  case FOwner.Kind of
    xdbdtRawData,
    xdbdtReal,
    xdbdtDateTime  :begin
                     GetMem(FRaw, 8);
                     PDateTime(FRaw)^:=ADateTime;
                   end;

    xdbdtInteger:begin
                     GetMem(FRaw, 8);
                     PInt64(FRaw)^:=round(ADateTime);
                   end;
    xdbdtBoolean  :begin
                     GetMem(FRaw, 1);
                     FAllocated:=1;
                     PByte(FRaw)^:=255;
                     if ADateTime = 0 then PByte(FRaw)^:=0;
                   end;
    xdbdtCurrency :begin
                     GetMem(FRaw, 8);
                     PCurrency(FRaw)^:=ADateTime;
                   end;
    xdbdtText     :begin
                     if FOwner.OutputMask = '' then pomw:=DateTimeToStr(ADateTime)
                        else pomw:=FormatDateTime(FOwner.OutputMask, ADateTime);
                     FAllocated:=Length(pomw) shl 1;
                     Move(pointer(pomw)^, FRaw^, FAllocated);
                     pomw:='';
                   end;
  end;

  ParseData;
  FOwner.Owner.Owner.Owner.SetLastError(xdbrcOK);
  if Assigned(FOnWritten) then FOnWritten(Self);
end;

function TXDBDataObj.GetWide:WideString;
begin
  if Assigned(FOnDataDemand) then FOnDataDemand(Self);

  result:='';
  if FRaw = nil then
{     begin
       if  FOwner.InitialValue<>Self then result:=FOwner.InitialValue.AsWideString;}
       exit;
{     end;}

  FOwner.Owner.Owner.Owner.SetLastError(xdbrcInvalidData);
  case FOwner.Kind of
     xdbdtRawData  :;
     xdbdtCurrency :begin
                      if FAllocated < 8 then Abort;
                      if FOwner.OutputMask = '' then result:=CurrToStr(PCurrency(FRaw)^)
                         else result:=FormatCurr(FOwner.OutputMask,PCurrency(FRaw)^);
                    end;
     xdbdtReal     :begin
                      if FAllocated < 8 then Abort;
                      if FOwner.OutputMask = '' then result:=FloatToStr(PDouble(FRaw)^)
                         else result:=FormatFloat(FOwner.OutputMask,PDouble(FRaw)^);
                    end;
     xdbdtDateTime :begin
                      if FAllocated < 8 then Abort;
                      if FOwner.OutputMask = '' then result:=DateTimeToStr(PDateTime(FRaw)^)
                         else result:=FormatDateTime(FOwner.OutputMask,PDateTime(FRaw)^);
                    end;
     xdbdtInteger  :begin
                      if FAllocated < 8 then Abort;
                      if FOwner.OutputMask = '' then result:=IntToStr(PInt64(FRaw)^)
                         else result:=FormatFloat(FOwner.OutputMask,PInt64(FRaw)^);
                    end;
     xdbdtBoolean  :begin
                      if PByte(FRaw)^ = 0 then result:=FOwner.Owner.Owner.Owner.NoString
                         else result:=FOwner.Owner.Owner.Owner.YesString;
                    end;
     xdbdtText     :begin
                      SetLength(result, FAllocated shr 1);
                      Move(FRaw^, pointer(result)^, FAllocated);
                    end;
  else
    Abort;
  end;
  FOwner.Owner.Owner.Owner.SetLastError(xdbrcOK);
end;

procedure TXDBDataObj.SetWide(AWide:WideString);
begin
  if FRaw<>nil then FreeMem(FRaw);

 if (KillSpaces(AWide) = '') and (FOwner.Kind <> xdbdtText) then
     begin
       FAllocated:=0;
       FRaw:=nil;
       exit;
     end;

  FAllocated:=8;
  case FOwner.Kind of
    xdbdtRawData,
    xdbdtText     :begin
                     FAllocated:=Length(AWide) shl 1;
                     GetMem(FRaw, FAllocated);
                     Move(pointer(AWide)^, FRaw^, FAllocated);
                   end;

    xdbdtCurrency :begin
                     GetMem(FRaw, 8);
                     PCurrency(FRaw)^:=StrToCurr(KillSpaces(AWide));
                   end;
    xdbdtReal:    begin
                     GetMem(FRaw, 8);
                     PDouble(FRaw)^:=StrToFloat(KillSpaces(AWide));
                   end;
    xdbdtDateTime  :begin
                     GetMem(FRaw, 8);
                     PDateTime(FRaw)^:=StrToDateTime(KillSpaces(AWide));
                   end;
    xdbdtInteger:begin
                     GetMem(FRaw, 8);
                     PInt64(FRaw)^:=StrToInt64(KillSpaces(AWide));
                   end;
    xdbdtBoolean  :begin
                     GetMem(FRaw, 1);
                     FAllocated:=1;
                     PByte(FRaw)^:=0;
                     if UpperCase(AWide)[1] <>
                        UpperCase(FOwner.Owner.Owner.Owner.NoString)[1]
                        then PByte(FRaw)^:=255;
                   end;
  end;

  ParseData;
  FOwner.Owner.Owner.Owner.SetLastError(xdbrcOK);
  if Assigned(FOnWritten) then FOnWritten(Self);
end;

function TXDBDataObj.DataSize:integer;
begin
  if Assigned(FOnDataDemand) then FOnDataDemand(Self);

  result:=FAllocated;
end;

procedure TXDBDataObj.ParseData;
begin
  if not Parse  then exit;

  //asa first test for min value
  with FOwner do
    if MinValue<>Self then
      case Kind of
        xdbdtInteger :if AsInt64 < MinValue.AsInt64 then AsInt64:=MinValue.AsInt64;
        xdbdtReal    :if AsDouble < MinValue.AsDouble then AsDouble:=MinValue.AsDouble;
        xdbdtBoolean :if AsBoolean < MinValue.AsBoolean then AsBoolean:=MinValue.AsBoolean;
        xdbdtCurrency:if AsCurrency < MinValue.AsCurrency then AsCurrency:=MinValue.AsCurrency;
        xdbdtDateTime:if AsDateTime < MinValue.AsDateTime then AsDateTime:=MinValue.AsDateTime;
        xdbdtText    :if CompareStringW(LOCALE_USER_DEFAULT,
                                      0,
                                      PWideChar(AsWideString),
                                      -1,
                                      PWideChar(MinValue.AsWideString),
                                      -1) = 1 then AsWideString:=MinValue.AsWideString;
  end;

  //and now for max value
  with FOwner do
    if MaxValue<>Self then
      case Kind of
        xdbdtInteger :if AsInt64 > MaxValue.AsInt64 then AsInt64:=MaxValue.AsInt64;
        xdbdtReal    :if AsDouble > MaxValue.AsDouble then AsDouble:=MaxValue.AsDouble;
        xdbdtBoolean :if AsBoolean > MaxValue.AsBoolean then AsBoolean:=MaxValue.AsBoolean;
        xdbdtCurrency:if AsCurrency > MaxValue.AsCurrency then AsCurrency:=MaxValue.AsCurrency;
        xdbdtDateTime:if AsDateTime > MaxValue.AsDateTime then AsDateTime:=MaxValue.AsDateTime;
        xdbdtText    :if CompareStringW(LOCALE_USER_DEFAULT,
                                      0,
                                      PWideChar(AsWideString),
                                      -1,
                                      PWideChar(MaxValue.AsWideString),
                                      -1) = 3 then AsWideString:=MaxValue.AsWideString;
  end;
end;

procedure TXDBDataObj.Assign(AData:TXDBDataObject);
begin
  if AData = Self then
     begin
       FOwner.Owner.Owner.Owner.SetLastError(xdbrcCannotAssignSelf);
       Abort;
       exit;
     end;


  if AData = nil then   //we are asked to empty our buffer
     begin              //we must do it in this way to keep our data safe
        FOwner.Owner.Owner.Owner.SetLastError(xdbrcUnexpected);
        if FRaw <> nil then FreeMem(FRaw);
        FRaw:=nil;              //sanity check
        FAllocated:=0;
        FOwner.Owner.Owner.Owner.SetLastError(xdbrcOK);
        if Assigned(FOnWritten) then FOnWritten(Self);
        exit;
     end;


  FOwner.Owner.Owner.Owner.SetLastError(xdbrcInvalidData);

  //get data size
  FAllocated:=AData.DataSize;    //also tests AData validity
  if FRaw <> nil then FreeMem(FRaw); //empty buffer
  //alloc memory
  GetMem(FRaw, FAllocated);
  //copy data
  Move(AData.PureRawData^, FRaw^, FAllocated);

  FOwner.Owner.Owner.Owner.SetLastError(xdbrcOK);
  if Assigned(FOnWritten) then FOnWritten(Self);
end;

end.
