//Part of XDB Core source code.
//Version 1.4; Revision 1
//Created by Tomas Koutny.
//for more info see http://www.rawos.cz/xdbe/index.html


//implementation of object representing database
unit FileLevel;

interface

uses EngineLevel, AbstractLayer, Stream64, XDBStructs, UserLevel, TableLevel;

type
    TXDBFile = class(TAXDBase)
               private
//                 FDestroyingSelf:Boolean;     //true, when destructor call
                 function TablePos(TblName:WideString):integer; 
                        //returns position of table in FTable
                        //if no such table returns position where it should be
                        //incremented by one and this number is multiplied by -1
                        //result:=-1*(DesiredPosition + 1);  
               protected
                 FTables:array of TXDBTbl; //for perfomance improvement 
                                           //this array is sorted in ascending
                                           //order by tables' names
                 FSpaces:TXDBPlacements;   //loaded free spaces in db file
                 FFreeedSpaces:TXDBPlacements; //freeed spaces in db file mista uvolnena v prubehu zpracovani
                        //for security reasons can not be performed write
                        //to this places (places in FReeedSpaces)
                        //for optimalization reasons FFreeedSpaces stores
                        //places containing links to free spaces, table, etc.

//                 FileStream:TFileStream64;
                 procedure LoadFile(FName:String); //loads db file
                 function GetTable(index:integer):TXDBTable; override;
                 function GetNamedTable(TblName:WideString):TXDBTable; override;
                 function CreateTable(TblName:WideString):TAXDBTable;
                        //creates new empty table named TblName a places
                        //it into FTable at correct position
                        //if TblName already exists returns nil
                 function AllocFreeSpace(HowMuch:Int64):Int64;
                        //allocs space in db file where can be written HowMuch
                        //bytes and returns its positions
                        //operates over FSpaces or returns end of file
                        //if HowMuch = 0 then returns 0
                 function AllocFreeSpaceFragmented(HowMuch:Int64):TXDBPlacements;
                        //has the same function like AllocFreeSpace but here
                        //returns list of free places as sum of their sizes
                        //is equal to HowMuch
                 procedure RegisterSpace(var At, What:TXDBPlacements;
                                             FreeAfterReg:boolean = false);
                        //join What at At, what can be nil, if FreeAfterReg
                        //then frees What from memory
                 procedure WritePlacements(Plcmnts:TXDBPlacements; At:Int64;
                           RegisterIt:Boolean = true);
                        //writes placements into FileStream at position At+8
                        //at position AT writes Length(Plcmnts) as Int64
                        //if RegisterIt then registers Plcmnts at FFreeedSpaces
                 function ReadPlacements(At:Int64; RegisterIt:Boolean = true):TXDBPlacements;                        
                        //loads from FileStream from position At block
                        //with positions into newly allocated memory
                        //if At = 0 returns nil
                        //if RegisterIt then registers result into FFreeedSpaces
                procedure FreeRecordPart(At:Int64);
                           //at position At frees part of record
                           //keep position(At) in stream
                           //At is identified by ID of record's part
               public
                 constructor Create(AOwner:TAXEngine; DBName:String;
                                    CreateIt:Boolean); override;
                 procedure Flush; override;
                 function TablesCount:Integer; override;
                 function AddTable(TblName:WideString):TXDBTable; override;
                 function TableRenameHandler(Sender:TAXDBTable; NewName:WideString):Boolean; override;
                        //handler for message of table rename
                        //funtion must be called by table => public declaration
                 procedure TableDestroyHandler(Sender:TAXDBTable); override;
                        //handler for message of table destroy
                 function PopPoss(FromPos, Count:Int64):TPositions; override;
                        //loads from db file table of offsets, there is
                        //count offsets and the place in db file will be marked
                        //freeed and free since next db save
                 function PushPoss(Poss:TPositions):Int64; override;
                        //saves into the db file table of offsets without
                        //any info about their count
                        //returns offset where table was saved
                        //place in db file will marked as at PopPoss
                        //Poss will not be freed from memory
                 procedure LoadData(FromPos, ID:Int64; Data:TAXDBDataObject); override;
                        //loads from db file part of record; part belongs
                        //to columnd ID, number of record is determined
                        //by position stored in db table object
                 procedure StoreData(var AtPos:Int64; ID:Int64; Data:TAXDBDataObject); override;
                        //tries to save Data at position AtPos, in way when
                        //data will belong to column ID
                        //returns position where the data has been written
                 procedure FreeBlock(From, Size:Int64); override; //uvedene misto
                        //marks place in db file at position From with size Size
                        //as free
                 procedure FreeRecord(AtPos:Int64); override; //misto zabrane
                        //frees record at position AtPos and marks place
                        //in db file as free


                 procedure Assimilate(Who:TXDBase); override;
                        //imports data from Who into self
                        //existing values will be replaced by imported ones
                        //or they'll be kept unchanged
                        //min, max & init values of tables' columns will remain
                        //unchanged
                        //no data conversion will be done                        


                 destructor Destroy; override;
               end;

implementation

uses SysUtils, windows, math, XDBRetCodes, ColLevel, UCol;

constructor TXDBFile.Create(AOwner:TAXEngine; DBName:String; CreateIt:Boolean);
var flag:DWORD;
    existed:boolean;
begin
  FOwner:=AOwner;
  FDestroyingSelf:=false;

  existed:=FileExists(DBName);
  if CreateIt and  existed and
     not(SysUtils.DeleteFile(DBName)) then Abort; //nelze-li vytvorit soubor

  //we will create our file
  flag:=OPEN_ALWAYS;
  if CreateIt then Flag:=CREATE_ALWAYS;
  FileStream:=TFileStream64.CreateW32(DBName,
                                      GENERIC_READ or GENERIC_WRITE,
                                      0,
                                      flag);
  //cache it
  FileStream:=FOwner.CreateCacheStream(FileStream);

  //and empty database
  FTables:=nil;
  FSpaces:=nil;
  FFreeedSpaces:=nil;

  //if neeeded then load data
  if not(CreateIt) and existed then LoadFile(DBName)
    else Flush; //else write header

  FOwner.SetLastError(xdbrcOK);
end;

destructor TXDBFile.Destroy;
var i:integer;
begin
  Flush; //data security reasons

  FDestroyingSelf:=true;
  for i:=0 to High(FTables) do
    FTables[i].Free;
  FTables:=nil;

  FileStream.Free;

  inherited;
end;

procedure TXDBFile.LoadFile(FName:String);
var hdr:TXDBHeader;
    plcs:TXDBPlacements;
    i,j,k,l,m:integer;
    pomTbl:TXDBTbl;
begin
  FOwner.SetLastError(xdbrcOK);

  //as first we have to read database header
  with FileStream do
     begin
       Position:=0;
       Read(hdr, SizeOf(TXDBHeader));
     end;

  //check for database format
  with hdr do
     if (Signature <> XDBFileSignature) or
        (MakeLong(LoVersion, HiVersion) > 7)  then
        begin
          FOwner.SetLastError(xdbrcUnknownFileFormat);
          Abort;
          exit;
        end;

  FSpaces:=nil;
  FFreeedSpaces:=nil;

  //DB header loooks OK => load tables' definitions
  plcs:=ReadPlacements(hdr.Tables);
  //we have positions => load definitions

  //we must sort tables to get 'em in ascending order by their names
  j:=High(plcs);
//  SetLength(FTables, j+1);
  FTables:=nil;
  for i:=0 to j do
     begin
       pomTbl:=TXDBTbl.Create(Self);
       pomTbl.ReadDefFromFile(plcs[i].Position, FileStream);
       k:=TablePos(pomTbl.Name);
       if k<0 then //we will not load duplicate tables
         begin
           k:=-1*(k+1);
           l:=Length(FTables);
           SetLength(FTables, l+1);
           for m:=k+1 to l do
             FTables[m]:=FTables[m-1];

           FTables[k]:=pomTbl;
         end else
            pomTbl.Free;
     end;

  //register as freeed space and free memory
  RegisterSpace(FFreeedSpaces, plcs, true);

  //load free spaces
  plcs:=ReadPlacements(hdr.Spaces);
  for i:=0 to High(plcs) do
     FreeBlock(plcs[i].Position, plcs[i].Size);
  plcs:=nil;
end;

procedure TXDBFile.Flush;
var XDBHeader:TXDBHeader;
    i,j:integer;
    plcmnts,
    oldFreeed:TXDBPlacements;

begin
  FOwner.SetLastError(xdbrcOK);

  //prepare db file header
  with XDBHeader do
     begin
       Move(XDBFileSignature, Signature, 12);
       loVersion:=7;
       hiVersion:=0;
       Tables:=0;
       Indexes:=0;
       Queries:=0;
       DTypes:=0;
       Spaces:=0;
       Specs:=0;
       Info:=0;
     end;
  //conditionally write it
  i:=Sizeof(TXDBHeader);
  with FileStream do
    begin
      Position:=0;
      if Size < i then
        begin
          Write(XDBHeader, i);
          exit;
        end else
            Read(XDBHeader, i); //keep old values
    end;

  //old freeed spaces will be written as free => new list of freeed spaces
  oldFreeed:=FFreeedSpaces;
  FFreeedSpaces:=nil;

  //write tables' definitions
  j:=Length(FTables);
  Plcmnts:=nil;
  SetLength(Plcmnts, j);
  for i:=0 to j-1 do
     begin
       //get size of table's definition
       Plcmnts[i].Size:=FTables[i].DefinitionSize;
       //find place to write table's definition
       Plcmnts[i].Position:=AllocFreeSpace(Plcmnts[i].Size);
       //we have it => write
       FTables[i].FlushDefIntoFile(Plcmnts[i].Position, FileStream);
     end;

  //we have tables' definitions => write their placements into file
  i:=Length(Plcmnts)*sizeof(TXDBPlacement);
  if i>0 then
    begin
      XDBHeader.Tables:=AllocFreeSpace(Int64(i+8));
      WritePlacements(Plcmnts, XDBHeader.Tables);
      //registr as now freeed space and free memory
      RegisterSpace(FFreeedSpaces, Plcmnts, true);
    end else
        XDBHeader.Tables:=0;

  //now we need to adjust list of free spaces and write 'em into the file
  //as first we have to know where we can write 'em
  i:=Length(FSpaces)+Length(oldFreeed);
  if i>0 then
     begin
        //sjednotime volna mista
        //unite free spaces
        RegisterSpace(FSpaces, oldFreeed, true);
        //write 'em into the file
        i:=Length(FSpaces)*SizeOf(TXDBPlacement);
        XDBHeader.Spaces:=AllocFreeSpace(Int64(i+8));
        WritePlacements(FSpaces, XDBHeader.Spaces);
      end else
          XDBHeader.Spaces:=0;

  //at the beginning of the file must be written db file header
  with FileStream do
     begin
       Position:=0;
       WriteBuffer(XDBHeader, sizeof(TXDBHeader));
     end;

  //force system to flush it
  FileStream.Flush;
end;


function TXDBFile.TablesCount:Integer;
begin
  result:=Length(FTables);
end;

function TXDBFile.TablePos(TblName:WideString):integer;
var LoBound, HiBound:Integer;
begin
  //we will use method of binary seeking
  //as first setup indexes how to search
  LoBound:=0;
  HiBound:=High(FTables);
  if HiBound = -1 then
     begin      //=> no table in evidence => new one will be first
       result:=-1;
       exit;
     end;

  repeat
    //find position where to have a look first
    result:=(HiBound+LoBound) shr 1;
    //compare it with taken of local settings isn't needed
    if FTables[result].Name = TblName then exit;

    //if we know the table then it lies above or below
    if FTables[result].Name < TblName then LoBound:=min(HiBound, integer(result+1))
       else HiBound:=max(LoBound, integer(result-1));

  until LoBound = HiBound;
  result:=LoBound;
  if FTables[LoBound].Name = TblName then exit;
  result:=1;
  //check for case of append
  if (HiBound = High(FTables)) and (FTables[HiBound].Name < TblName) then inc(result);

  //no such table => return where it should be
  result:=-1*(LoBound+result);
end;

function TXDBFile.GetTable(index:integer):TXDBTable;
begin
  result:=nil;
  if (index < 0) or (index >= Length(FTables)) then exit;
  result:=FTables[Index];
end;

function TXDBFile.GetNamedTable(TblName:WideString):TXDBTable;
begin
  result:=Tables[TablePos(TblName)];
end;


function TXDBFile.CreateTable(TblName:WideString):TAXDBTable;
var i, j, k:integer;
begin
  //as first have a look if table already exists
  i:=TablePos(TblName);
  if i>=0 then
     begin //table already exists => nothing to create
       result:=nil;
       FOwner.SetLastError(xdbrcTableAlreadyExists);
       Abort;
       exit;
     end;

  //we have to create new table
  result:=TXDBTbl.Create(Self);
  result.Name:=TblName;
  //put table into FTables
  j:=Length(FTables);
  SetLength(FTables, j+1);
  for k:=-i to j do
      FTables[k]:=FTables[k-1];
  FTables[-1*(i+1)]:=result as TXDBTbl;
end;

function TXDBFile.AddTable(TblName:WideString):TXDBTable;
begin
  FOwner.SetLastError(xdbrcUnableToCreateTable);
  //we can get CS:EIP into exception with already set Flag
  //=> now we set default
  try
    result:=CreateTable(TblName);
    FOwner.SetLastError(xdbrcOK);
  except
    result:=nil;
  end;
end;

function TXDBFile.AllocFreeSpace(HowMuch:Int64):Int64;
var i,j,k:integer;
    pom:int64;
begin
  if HowMuch = 0 then
     begin
       result:=0;
       exit;
     end;

  //search through list of free spaces
  for i:=0 to High(FSpaces) do
    if FSpaces[i].Size >= HowMuch then
       begin //we found place that we need
         result:=FSpaces[i].Position;
         //check for sizes, if they're equal or not
         if HowMuch = FSpaces[i].Size  then
            begin    //remove gap from list
              k:=High(FSpaces);
              for j:=i to k-1 do
                 FSpaces[j]:=FSpaces[j+1];
              SetLength(FSpaces, k);
            end else
                begin //there's some free space left
                  FSpaces[i].Size:=FSpaces[i].Size-HowMuch;
                  FSpaces[i].Position:=FSpaces[i].Position+HowMuch;
                end;
         exit;
       end;

  //nothing found => return end of file
  i:=High(FSpaces);
  pom:=FileStream.Size; //try to include last fragment
  if (i>=0) and (FSpaces[i].Position+FSpaces[i].Size = pom) then
     begin
       result:=FSpaces[i].Position;
       SetLength(FSpaces, i);
     end else
        result:=pom;
end;

procedure TXDBFile.RegisterSpace(var At, What:TXDBPlacements;
                                 FreeAfterReg:boolean = false);
var i,j,k:integer;
begin
  if What = nil then exit;

{  j:=Length(At);
  k:=Length(What);

  SetLength(At, j+k);
  for i:=j to j+k-1 do
     At[i]:=What[i-j];
}

  for i:=0 to High(What) do
     begin
       if (What[i].Size = 0) or
          (What[i].Position < 72) then continue;       //error check
       
       //as first we have to find in At position where What[i] should be
       k:=Length(At);
       j:=0;
       while (j<k) and (What[i].Position>At[j].Position+At[j].Size-1) do
          inc(j);

       if j = k then
          begin  //place will get appended
            SetLength(At, k+1);
            At[k]:=What[i];

            continue; //no more continuing in this cycle
          end;

       //we have positions where What[i] should be, but before we create new
       //block in At we will have a look for possibility if What[i] can be
       //merged with another existing block in At
       if What[i].Position+What[i].Size<At[j].Position then
          begin  //no such possibility, record will placed before block
            SetLength(At, k+1);
            for k:=High(At) downto j+1 do
                At[k]:=At[k-1];

            At[j]:=What[i];

            continue; //no more continuing in this cycle
          end;

       //now block will be enlarged
       if What[i].Position >= At[j].Position then
          At[j].Size:=max(At[j].Size, What[i].Position+What[i].Size-At[j].Position)
          else At[j].Size:=max(At[j].Position+At[j].Size-What[i].Position, What[i].Size);
       At[j].Position:=min(At[j].Position, What[i].Position);

       //block adjusted => check if it interfere with previous block
       if (j = 0) or (At[j].Position>At[j-1].Position+At[j-1].Size) then continue; //no
       //yes, join
       dec(j);
       At[j].Size:=At[j+1].Position+At[j+1].Size-At[j].Position+1;
       //and adjust At
       for k:=j+1 to High(At)-1 do
          At[k]:=At[k+1];
       SetLength(At, High(At));

       //check if block interfere with next block
       k:=Length(At);
       if (j = k-1) or (At[j].Size+At[j].Position < At[j+1].Position) then
           continue; //no

       //yes, merge next block with current block
       At[j].Size:=At[j+1].Position+At[j+1].Size-At[j].Position+1;
       //and adjust At
       for k:=j+1 to High(At)-1 do
          At[k]:=At[k+1];
       SetLength(At, High(At));
     end;

  if FreeAfterReg then What:=nil;
end;


procedure TXDBFile.WritePlacements(Plcmnts:TXDBPlacements; At:Int64;
                                                     RegisterIt:Boolean = true);
var pom:Int64;
    plc:TXDBPlacements;
begin
  pom:=Length(Plcmnts);

  with FileStream do
    begin
      Position:=At;
      Write(pom, sizeof(pom));
      if pom > 0 then Write(pointer(Plcmnts)^, pom*Sizeof(TXDBPlacement));
    end;


  if RegisterIt and (pom>0) then
     begin
        SetLength(plc, 1);
        plc[0].Position:=At;
        plc[0].Size:=pom*SizeOf(TXDBPlacement)+8;
        RegisterSpace(FFreeedSpaces, plc, true);
     end;
end;

function TXDBFile.ReadPlacements(At:Int64; RegisterIt:Boolean = true):TXDBPlacements;
var pom:Int64;
    plc:TXDBPlacements;
begin
  result:=nil;
  if At = 0 then exit;

  with FileStream do
    begin
      Position:=At;
      Read(pom, SizeOf(pom));
      SetLength(result, pom);
      if pom >0 then Read(pointer(result)^, pom*SizeOf(TXDBPlacement));
    end;

  if RegisterIt and (pom>0) then
     begin
        SetLength(plc, 1);
        plc[0].Position:=At;
        plc[0].Size:=pom*SizeOf(TXDBPlacement)+8;
        RegisterSpace(FFreeedSpaces, plc, true);
     end;
end;

function TXDBFile.TableRenameHandler(Sender:TAXDBTable; NewName:WideString):Boolean;
var newpos, oldpos, i:integer;
begin
  //if caller doesn't has valid name then we have to think caller is
  //in phase of creation
  if Sender.Name = ''then
    begin
      result:=true;
      exit;
    end;

  if NewName = '' then
    begin
      Owner.SetLastError(xdbrcInvalidName);
      result:=false;
      Abort;
      exit;
    end;

  //look for same name
  newpos:=TablePos(NewName);
  if newpos >= 0 then
     begin  //table already exists => cann't rename
       result:=false;
       FOwner.SetLastError(xdbrcTableAlreadyExists);
       Abort;
       exit;
     end;

  //table doesn't exists yet => reorder objects in FTables
  //we must also adjust newpos
  newpos:=-1*(newpos+1);
  oldpos:=TablePos(Sender.Name);
  if oldpos > newpos then
     begin
       inc(newpos); //no table to add
       for i:=oldpos downto newpos+1 do
          FTables[i]:=FTables[i-1]
     end else
        begin
          dec(newpos); //no table to add
          for i:=oldpos to newpos-1 do
              FTables[i]:=FTables[i+1];
        end;


  FTables[NewPos]:=Sender as TXDBTbl;

  //finally => table can be renamed
  result:=true;
  FOwner.SetLastError(xdbrcOK);
end;

procedure TXDBFile.TableDestroyHandler(Sender:TAXDBTable);
var tblpos:integer;
    i, j:integer;
begin
  if FDestroyingSelf then exit;

  //look for table's position
  tblpos:=TablePos(Sender.Name);
  if tblpos<0 then exit; //some error happened

  //reorder FTables
  j:=High(FTables);
  for i:=tblpos to j-1 do
     FTables[i]:=FTables[i+1];

  SetLength(FTables, j);
end;

function TXDBFile.PopPoss(FromPos, Count:Int64):TPositions;
var SizeToLoad:Int64;
    plc:TXDBPlacements;
begin
  result:=nil;
  if (Count = 0) or (FromPos = 0) then exit;

  //load data
  FileStream.Position:=FromPos;
  SizeToLoad:=Count shl 3;
  SetLength(result, Count);
  FileStream.Read(pointer(result)^, SizeToLoad);

  //and register place
  SetLength(plc, 1);
  plc[0].Position:=FromPos;
  plc[0].Size:=SizeToLoad;
  RegisterSpace(FFreeedSpaces, plc, true);
end;

function TXDBFile.PushPoss(Poss:TPositions):Int64;
var SizeToSave:Int64;
    plc:TXDBPlacements;
begin
  FOwner.SetLastError(xdbrcOK);
  result:=0;
  SizeToSave:=Length(poss) shl 3;
  if SizeToSave = 0 then exit;

  result:=AllocFreeSpace(SizeToSave);
  FileStream.Position:=result;
  FileStream.Write(pointer(Poss)^, SizeToSave);

  //and register place
  SetLength(plc, 1);
  plc[0].Position:=result;
  plc[0].Size:=SizeToSave;
  RegisterSpace(FFreeedSpaces, plc, true);
end;

procedure TXDBFile.LoadData(FromPos, ID:Int64; Data:TAXDBDataObject);
var plcs:TXDBPlacements;
    i, j:integer;
    pom:pointer;
begin
  FOwner.SetLastError(xdbrcOK);
  //as default action empty data buffer
  Data.AsBuffer[0]:=nil;

  //load info about record, because it has the same structure as info about
  //placement we will use plcs
  plcs:=ReadPlacements(FromPos, false);
  //we have to look if record contains desired part
  j:=-1;
  for i:=0 to High(plcs) do
      if plcs[i].Position = ID then       //position means ID of part of record
        begin
          j:=i;
          break;
        end;

  if (j = -1) or (plcs[j].Size = 0) then exit; //desired part is NULL
  //desired part is saved => we can load it
  j:=plcs[j].Size;
  plcs:=nil;
  plcs:=ReadPlacements(j, false); //size meand position with list of fragments
        //which compose part of record
  //find out size of part into j
  j:=0;
  for i:=0 to High(plcs) do
     j:=j+plcs[i].Size;
  //alloc needed memory
  Data.AsBuffer[j]:=nil;

  //we can staart loading now
  pom:=Data.PureRawData;  //offset into memory
  for i:=0 to High(plcs) do
     begin
       FileStream.Position:=plcs[i].Position;
       FileStream.Read(pom^, plcs[i].Size);
       pom:=pointer(integer(pom)+integer(plcs[i].Size));
     end;

  //finally we can free loaded positions
  plcs:=nil;
end;

procedure TXDBFile.StoreData(var AtPos:Int64; ID:Int64; Data:TAXDBDataObject);
var plcs, frags:TXDBPlacements;
    i, j, k:integer;
    pom, oldPos, oldSize:Int64;
    tmp:pointer;
begin
  FOwner.SetLastError(xdbrcOK);
  plcs:=nil;
  oldSize:=0;
  if AtPos <> 0 then
            //record has been already defined =>
            //load info about record
            //because it has the same structure as info about placements
            //we will use plcs
     begin
       plcs:=ReadPlacements(AtPos, false);
       oldSize:=8+(Length(plcs) shl 4);
     end;

  //we have to look if record contains part to be writtem
  j:=-1;
  for i:=0 to High(plcs) do
      if plcs[i].Position = ID then       //position means ID of part of record
        begin
          j:=i;
          break;
        end;

  if (j>=0) and (Data.DataSize = 0) then
      begin             //record exists but we have to cancel it
        //compute size of current array in file
        pom:=8+(Length(plcs) shl 4);
        //prepare new array
        k:=High(plcs);
        for i:=j to k-1 do
            plcs[i]:=plcs[i+1];
        SetLength(plcs, k);
        oldPos:=AtPos;  //this will be needed to free old block
        if k > 0 then   //if any record remains
           begin
             AtPos:=AllocFreeSpace(8+(k shl 4));
             WritePlacements(plcs, AtPos, false);
           end else
               AtPos:=0;
        //at the end we will free block in db file
        FreeBlock(oldPos, pom);
        //and end
        plcs:=nil;
        exit;
      end;

  if (Data.DataSize = 0) and (j = -1) then exit; //no record and nothing to save

  //we have to prepare locations of fragments
  frags:=AllocFreeSpaceFragmented(Data.DataSize);
  //write record
  tmp:=Data.PureRawData;  //offset into memory
  for i:=0 to High(Frags) do
      begin
        FileStream.Position:=frags[i].Position;
        FileStream.Write(tmp^, frags[i].Size);
        tmp:=pointer(integer(tmp)+integer(frags[i].Size));
      end;

  //write fragments
  pom:=8+(Length(frags) shl 4); //size of array with fragments
  oldPos:=AllocFreeSpace(pom);
  WritePlacements(frags, oldpos, false);

  if j = -1 then //we have to defined part of record
     begin
       k:=Length(plcs);
       SetLength(plcs, k+1);
       plcs[k].Position:=ID;
//       plcs[k].Position:=(Data.Owner as TXDBCol).ID;
       plcs[k].Size:=oldPos;  //we can not forget that here is another placement
     end else
        begin   //record already was there => only to change position
          k:=j;

          //free old blocks with record
//          FreeBlock(plcs[k].Position, plcs[k].Size);
          FreeRecordPart(plcs[k].Size);//we can not forget that here is another placement

          plcs[k].Position:=ID;
//          plcs[k].Position:=(Data.Owner as TXDBCol).ID;
          plcs[k].Size:=oldPos; //we can not forget that here is another placement
        end;

  //now we need to write new data's placement
  //free old
  if AtPos<>0 then FreeBlock(AtPos, oldSize);
  //and write new
  AtPos:=AllocFreeSpace(8+(Length(plcs) shl 4));
  WritePlacements(plcs, AtPos, false);
end;

procedure TXDBFile.FreeBlock(From, Size:Int64);
var plcs:TXDBPlacements;
begin
  SetLength(plcs, 1);
  plcs[0].Position:=From;
  plcs[0].Size:=Size;
  RegisterSpace(FSpaces, plcs, true);
end;

function TXDBFile.AllocFreeSpaceFragmented(HowMuch:Int64):TXDBPlacements;
var i,j,k:integer;
    pom:Int64;
begin
  result:=nil;
  if HowMuch = 0 then exit;
  i:=0;
  j:=Length(FSpaces);
  while (HowMuch > 0) and (i<j) do
    begin  //stepwise transfer free blocks into our list
      SetLength(result, i+1);
      result[i].Position:=FSpaces[i].Position;
      result[i].Size:=FSpaces[i].Size;

      HowMuch:=HowMuch-FSpaces[i].Size;
      if HowMuch < 0 then       //we have allocate needed space
        begin                   //only to adjust sizes and positions
          result[i].Size:=result[i].Size+HowMuch;
          FSpaces[i].Position:=FSpaces[i].Position+result[i].Size;
          FSpaces[i].Size:=-HowMuch;

          break;
        end;

      inc(i);
    end;

  //now in i-1 is id of recently freeed block
  if FSpaces <> nil then
     begin
       for k:=i to j-1 do
         FSpaces[k-i]:=FSpaces[k];

       //count new size of array of free blocks
       j:=j-i;
       SetLength(FSpaces, j);
     end;

  if HowMuch>0 then
     begin     //there were not enough spaces free => utilization of end of file
       j:=length(result);
       SetLength(result, j+1);
       pom:=FileStream.Size;
       i:=High(FSpaces);
       if i>=0 then
          begin                       //there is at least one free space
            if FSpaces[i].Position+FSpaces[i].Size <> pom then
                result[j].Position:=pom
                else begin
                       result[j].Position:=FSpaces[i].Position;
                       SetLength(FSpaces, i);
                     end;
          end else //no free spaces in file
              result[j].Position:=pom;  

       result[j].Size:=HowMuch;
     end;
end;

procedure TXDBFile.FreeRecord(AtPos:Int64);
var plcs:TXDBPlacements;
    i:integer;

begin
  if AtPos = 0 then exit; //record doesn't take any space
  //load record placement
  plcs:=ReadPlacements(AtPos, false);
  for i:=0 to High(plcs) do
     FreeRecordPart(plcs[i].Size);
  //free memory
  FreeBlock(AtPos, 8+(Length(plcs) shl 4));
  plcs:=nil;
end;

procedure TXDBFile.FreeRecordPart(At:Int64); //at position At frees part of record
                                             //keep position(At) in stream
var plcs:TXDBPlacements;
    i:integer;
    befSeek:int64;
begin
  befSeek:=FileStream.Position;

  plcs:=ReadPlacements(At, false);
  for i:=0 to High(plcs) do
     FreeBlock(plcs[i].Position, plcs[i].Size);
  //free memory
  FreeBlock(At, 8+(Length(plcs) shl 4));
  plcs:=nil;

  FileStream.Position:=befSeek;
end;

procedure TXDBFile.Assimilate(Who:TXDBase);
var i, j, k:integer;
    tbl, wtbl:TXDBTable;
    col:TXDBColumn;
    LookUp:array of integer;
begin
  if Who = Self then
     begin
       FOwner.SetLastError(xdbrcCannotAssimilateSelf);
       Abort;
       exit;
     end;

  FOwner.SetLastError(xdbrcUnexpected);
  for i:=0 to Who.TablesCount-1 do
     begin
       //check if table Who.Tables[i] exists in our database
       wtbl:=Who.Tables[i];
       try
         tbl:=GetNamedTable(wtbl.Name);
         if tbl = nil then Abort;
       except
         //no table or table is inaccessible => try to create it
         tbl:=AddTable(wtbl.Name);
         if tbl = nil then continue; //something is wrong with this table
       end;
       //we have table => adjust columns and then import data
       //there is no quarantee that columns will be placed at
       //same position => build look up table to make quick translation
       SetLength(LookUp, wtbl.ColumnsCount); //maximaly this count

       //check wtbl's columns
       for j:=0 to wtbl.ColumnsCount-1 do
         begin
           LookUp[j]:=tbl.NamedCol(wtbl.Columns[j].Name);
           if LookUp[j]<0 then  //column is not defined => create it
              begin
                col:=tbl.AddColumn(wtbl.Columns[j].Name);
                //column added => copy its settings
                col.Assign(wtbl.Columns[j]);

                //keep its position - we don't know where it has been added
                LookUp[j]:=tbl.NamedCol(col.Name);
              end;
         end;

       //columns checked => copy data
       //get enough of records
       j:=max(wtbl.Records, tbl.Records);
       while tbl.Records < j do
          tbl.CreateRecord(-1);
       //copy data
       for k:=0 to wtbl.ColumnsCount-1 do
          for j:=0 to wtbl.Records-1 do
             tbl[LookUp[k], j].Assign(tbl[k, j]);

       //free memory  
     end;

  //everything was OK
  FOwner.SetLastError(xdbrcOK);
end;

end.
