//Part of XDB Core source code.
//Version 1.2
//Created by Tomas Koutny.
//for more info see http://www.rawos.cz/xdbe/index.html


//object representing one table
unit TableLevel;

interface

uses UserLevel, AbstractLayer, stream64, ColLevel, DataObj, UCol;

type
    TXDBTbl = class(TAXDBTable)
              private
                FData:TXDBDataObj;       //main data object used to get table
                                         //data accesible
                FCol:TUniversalColumn;   //object allowing utilization of FData
                LRec:integer;            //recently referenced record's number
                LCol:TXDBCol;            //recently used column for reading
                FDataDemanded:Boolean;   //semaphor for FOnDataDemand
                FDestroyingSelf:boolean; //true, when destructor's call 
                FRecPoss:TPositions;     //record's offsets in the db file
                FDataPossLoaded:boolean; //true, when data has been properly
                             //loaded
                             //this flag is tested in destructor if there's
                             //need to remove data from db file
                FSavedAt:Int64; //position where is in db file stored
                        //table's definition; if zero then table is not stored
                        //utilized when changine FRecPoss, so we can loose
                        //maximaly definition of last (timely) record
                procedure OnDataWrittenHandler(Sender:TXDBDataObj);
                        //procedure that writes data into database
                procedure FOnDataDemand(Sender:TXDBDataObj); 
                        //procedure that reads data from database
                procedure FlushFRecPoss(At:Int64);
                        //writes location of table's records in db file at
                        //position At
                        //At must points to the header of table definition
              protected
                FOnRename:TXDBTableRename;
                FOnDestroy:TXDBTableDestroy;
                FCols:array of TXDBCol;
                function CanRename(NewName:WideString):boolean; 
                        //returns true if rename to NewName is possible
                        //calss FOnRename
                function GetColumn(index:integer):TXDBColumn; override;
                procedure FreeCols;   //free columns and their data from memory
                function GetData(ACol, ARec:integer):TXDBDataObject; override;
                function GetRecsCount:integer; override;
              public
                constructor Create(AOwner:TAXDBase); override;
                destructor Destroy; override;


                function NamedCol(ColName:WideString):integer; override;
                        //returns position of column named ColName
                        //if column doesn't exists returns negative value

                procedure SetName(NewName:WideString); override;

                function DefinitionSize:Int64;  //returns how much bytes will
                                                //be needed to store table's
                                                //definition in db file
                procedure FlushDefIntoFile(At:Int64; Str:TStream64);
                        //writes table's defitinition into strema at position At
                procedure ReadDefFromFile(At:Int64; Str:TStream64);
                        //loads table's definition from stream at position At

                function AddColumn(ColName:WideString):UserLevel.TXDBColumn; override;
                        //creates new column in table
                        //this values will be initialized
                        // Name: ColName
                        // Kind: RawDaa
                        //rest of values will be zeros or empty strings
                        //returns newly created column or nil(failure)

                procedure ColumnDestroyHandler(Sender:TAXDBColumn); override;
                function ColumnRenameHandler(Sender:TAXDBColumn;
                             NewName:WideString):boolean; override;

                procedure MoveColumn(FromPos, ToPos:integer); override;
                function ColumnsCount:integer; override;

                procedure CreateRecord(Index:integer); override;
                procedure DeleteRecord(index:integer); override;

                procedure MoveRecord(FromPos, ToPos:integer); override;
                procedure XCHGRecord(FromPos, ToPos:integer); override;
              end;


implementation

uses XDBStructs, SysUtils, XDBRetCodes, UserStructs, routines;

constructor TXDBTbl.Create(AOwner:TAXDBase);
begin
  FSavedAt:=0;
  FDataPossLoaded:=false;

  FDestroyingSelf:=false;
  FOwner:=AOwner;
  FTemp:=nil;
  FOnRename:=FOwner.TableRenameHandler;
  FOnDestroy:=FOwner.TableDestroyHandler;
  FCol:=TUniversalColumn.Create(Self);
  FData:=TXDBDataObj.Create(FCol);
  FData.OnDataWritten:=OnDataWrittenHandler;
  FData.OnDataDemand:=FOnDataDemand;
  FDataDemanded:=false;
  FRecPoss:=nil;
  LRec:=-1;
  LCol:=nil;

  FDataPossLoaded:=true;
end;

destructor TXDBTbl.Destroy;
var i:integer;
begin
  FDestroyingSelf:=true;
  if Assigned(FOnDestroy) then FOnDestroy(Self);

  if FOwner.DestroyingSelf and FDataPossLoaded then
     begin  //table is destroying => we have to destroy our records too
       for i:=0 to High(FRecPoss) do
           FOwner.FreeRecord(FRecPoss[i]);
     end;

  for i:=0 to High(FCols) do
    FCols[i].Free;
  FCols:=nil;

  FRecPoss:=nil;

  FData.Free;
  FCol.Free;

  inherited;
end;

procedure TXDBTbl.SetName(NewName:WideString);
begin
  FOwner.Owner.SetLastError(xdbrcCannotRenameTable);
  //ask for rename acknowledge
  if not CanRename(NewName) then
     begin //cann't be done => exception
       Abort;
       exit;
     end;

  FName:=NewName;
  FOwner.Owner.SetLastError(xdbrcOK);
end;

function TXDBTbl.DefinitionSize:Int64;
var i:integer;
begin
  result:=56+(Length(FName) shl 1);
  for i:=0 to High(FCols) do
    result:=result+FCols[i].DefinitionSize;
end;

procedure TXDBTbl.FlushDefIntoFile(At:Int64; Str:TStream64);
var pom:integer;
    pom2, AfterSeek:Int64;
begin
  Str.Position:=At;

  //accurate location of record can be written at the end
  Str.Write(pom2, 8);
  Str.Write(pom2, 8);

  pom2:=0;
  str.Write(pom2, 8); //flags - not used here

  //name
  WriteWideString(Str, FName);

  //PrimaryKey
  WriteWideString(Str, '');
  //SortedBy
  WriteWideString(Str, '');

  //columns' count
  pom2:=Length(FCols);
  str.Write(pom2, 8);

  //coulumns
  for pom:=0 to High(FCols) do
     FCols[pom].FlushDefIntoStream(Str);

  AfterSeek:=Str.Position;

  //count and locations of records
  Str.Position:=At;
  pom2:=Length(FRecPoss);
  str.Write(pom2, 8);
  pom2:=FOwner.PushPoss(FRecPoss);
  Str.Position:=At+Int64(8); //we have to return file pointer
  str.Write(pom2, 8);

  Str.Position:=AfterSeek;
  FSavedAt:=At;
end;

procedure TXDBTbl.ReadDefFromFile(At:Int64; Str:TStream64);
var pom:Int64;
    pomw:WideString;
    i:integer;
begin
  FDataPossLoaded:=false;

  Str.Position:=At;

  //records count
  Str.Read(pom, 8);
///  SetLength(FRecPoss, pom);
  i:=integer(pom);
  //position of firts record
  Str.Read(pom, 8);
  FRecPoss:=Owner.PopPoss(pom, i);

  Str.Position:=At+Int64(24); //skip unused data

  //name
  FName:=ReadWideString(Str);
  //primaryKey
  pomw:=ReadWideString(Str);
  pomw:='';
  //SortedBy
  pomw:=ReadWideString(Str);
  pomw:='';

  //columns' count
  Str.Read(pom, 8);
  i:=pom;
  FreeCols;
  SetLength(FCols, i);
  for i:=0 to High(FCols) do
     begin
       FCols[i]:=TXDBCol.Create(Self);
       FCols[i].ReadDefFromStream(Str);
     end;

  FDataPossLoaded:=true;
  FSavedAt:=At;
end;

procedure TXDBTbl.FlushFRecPoss(At:Int64);
var pom2:Int64;
    osize, oplc:Int64;
begin
  if At = 0 then exit;

  with FOwner do
    begin
      FStream.Position:=At;
      //we will read location of old block with positions
      FStream.Read(osize, 8);
      FStream.Read(oplc, 8);

      //and we have to write new positions
      FStream.Position:=At;
      pom2:=Length(FRecPoss);
      FStream.Write(pom2, 8);
      pom2:=PushPoss(FRecPoss);
      FStream.Position:=At+Int64(8); //file pointer return
      FStream.Write(pom2, 8);

      FSavedAt:=At;

      //finally we will old block
      FreeBlock(oplc, osize shl 3);
    end;
end;

function TXDBTbl.CanRename(NewName:WideString):boolean;
begin
  result:=not(Assigned(FOnRename)) or FOnRename(Self, NewName);
end;

function TXDBTbl.AddColumn(ColName:WideString):UserLevel.TXDBColumn;
var j:integer;
begin
  FOwner.Owner.SetLastError(xdbrcOK);
  //verification that there's no column with ColName
  if NamedCol(ColName) >=0 then
     begin
       FOwner.Owner.SetLastError(xdbrcColAlreadyExists);
       result:=nil;
       Abort;
       exit;
     end;


  result:=TXDBCol.Create(Self);
  result.Name:=ColName;
  with result do
    begin
       Kind:=xdbdtRawData;
       Comment:='';
       InputMask:='';
       OutputMask:='';
       MinLength:=0;
       MaxLength:=0;
    end;

  j:=Length(FCols);
  SetLength(FCols, j+1);
  FCols[j]:=result as TXDBCol;

  //we have to rebuild db structure by flush of everything except records
  FOwner.Flush;
end;

function TXDBTbl.GetColumn(index:integer):UserLevel.TXDBColumn;
begin
  FOwner.Owner.SetLastError(xdbrcColNotFound);
  result:=nil;
  if (index <0) or (index>High(FCols)) then exit;
  FOwner.Owner.SetLastError(xdbrcOK);
  result:=FCols[index];
end;

procedure TXDBTbl.ColumnDestroyHandler(Sender:TAXDBColumn);
var i,j,k:integer;
begin
  if FDestroyingSelf then exit;

  //find column to destroy
  j:=High(FCols);
  for i:=0 to j-1 do
     if FCols[i] = Sender then break;

  if FCols[i] <> Sender then exit; //security check

  //zrusime casti zaznamu
  //destroy record's parts
  for k:=0 to High(FRecPoss) do
      Data[i, k].AsBuffer[0]:=nil;

  //final column free action
  for k:=i to j-1 do
    FCols[k]:=FCols[k+1];

  SetLength(FCols, j);

  //and rebuild
  FOwner.Flush;
end;

procedure TXDBTbl.MoveColumn(FromPos, ToPos:integer);
var pom:TXDBCol;
    i:integer;    
begin
  if FromPos>ToPos then
    begin
      if ToPos<0 then ToPos:=0;
      i:=High(FCols);
      if FromPos>i then FromPos:=i;
      if ToPos = FromPos then exit;

      pom:=FCols[FromPos];                 //backup of object to move

      for i:=FromPos downto ToPos+1 do
          FCols[i]:=FCols[i-1]
    end else
        begin
          if FromPos<0 then FromPos:=0;
          i:=High(FCols);
          if ToPos>i then ToPos:=i;
          if ToPos = FromPos then exit;

          pom:=FCols[FromPos];                 //backup of object to move

          for i:=FromPos to ToPos-1 do
               FCols[i]:=FCols[i+1];
        end;

  FCols[ToPos]:=pom;       //restoration of moved object
end;

function TXDBTbl.ColumnsCount:integer;
begin
  result:=Length(FCols);
end;

function TXDBTbl.NamedCol(ColName:WideString):integer;
var j:integer;
begin
  result:=-1;
  for j:=0 to High(FCols) do
    if ColName = FCols[j].Name then
        begin
          result:=j;
          exit;
        end;
end;

function TXDBTbl.ColumnRenameHandler(Sender:TAXDBColumn;NewName:WideString):boolean;
var i:integer;
begin
  result:=true;
  i:=NamedCol(NewName);
  if i>=0 then result:=FCols[i] = Sender;
end;

procedure TXDBTbl.FreeCols;
begin
  while FCols <> nil do
    FCols[0].Free;
end;

procedure TXDBTbl.OnDataWrittenHandler(Sender:TXDBDataObj);
begin
  FDataDemanded:=true; //in this time we have new data in memory
  FOwner.StoreData(FRecPoss[LRec], FCol.ID, FData);
end;

function TXDBTbl.GetData(ACol, ARec:integer):TXDBDataObject;
begin
  if (FCols = nil) or (FrecPoss = nil) then
     begin
       result:=nil;
       exit;
     end;

  result:=FData; //here will be always read
  //set simulated column
  FCol.Simulated:=FCols[ACol];
  if (LCol = FCol.Simulated) and (LRec = ARec) then exit; //already loaded
  //rememeber item
  LCol:=FCol.Simulated;
  LRec:=ARec;
  //and that its needs to be loaded before read
  FDataDemanded:=false;
end;

procedure TXDBTbl.FOnDataDemand(Sender:TXDBDataObj);
begin
  if FDataDemanded then exit;
  FOwner.LoadData(FRecPoss[LRec], LCol.ID, Sender);
  FDataDemanded:=true;
end;

function TXDBTbl.GetRecsCount:integer;
begin
  result:=Length(FRecPoss);
end;

procedure TXDBTbl.CreateRecord(Index:integer);
var i,j:integer;
begin
  if Length(FCols) <= 0 then
     begin
       FOwner.Owner.SetLastError(xdbrcNoColumns);
       Abort;
       exit;
     end;

  FOwner.Owner.SetLastError(xdbrcNotEnoughMemory);
  i:=Length(FRecPoss);
  if Index<0 then
     begin  //we will make some append
       SetLength(FRecPoss, i+1);
       FRecPoss[i]:=0; //new empty record

       FlushFRecPoss(FSavedAt);
       FOwner.Owner.SetLastError(xdbrcOK);
       exit;
     end;

  //bounds check
  if Index>=i then
     begin
       FOwner.Owner.SetLastError(xdbrcInvalidIndex);
       Abort;
       exit;
     end;

  SetLength(FRecPoss, i+1);
  for j:=i downto index+1 do
    FRecPoss[j]:=FRecPoss[j-1];
  FRecPoss[index]:=0;

  FlushFRecPoss(FSavedAt);
  FOwner.Owner.SetLastError(xdbrcOK);
end;

procedure TXDBTbl.DeleteRecord(index:integer);
var i,j:integer;
    pom:Int64;
begin
  FOwner.Owner.SetLastError(xdbrcInvalidIndex);
  i:=High(FRecPoss);
  //bounds check
  if (index<0) or (index>i) then
      begin
        Abort;
        exit;
      end;

  pom:=FRecPoss[index];
  for j:=index to i-1 do
      FRecPoss[j]:=FRecPoss[j+1];

  FOwner.FreeRecord(pom);
  SetLength(FRecPoss, i);

  FlushFRecPoss(FSavedAt);
  FOwner.Owner.SetLastError(xdbrcOK);
end;

procedure TXDBTbl.MoveRecord(FromPos, ToPos:integer);
var pom:Int64;
    i:integer;
begin
  if FromPos>ToPos then
    begin
      if ToPos<0 then ToPos:=0;
      i:=High(FRecPoss);
      if FromPos>i then FromPos:=i;
      if ToPos = FromPos then exit;

      pom:=FRecPoss[FromPos];                 //backup of object to move

      for i:=FromPos downto ToPos+1 do
          FRecPoss[i]:=FRecPoss[i-1]
    end else
        begin
          if FromPos<0 then FromPos:=0;
          i:=High(FRecPoss);
          if ToPos>i then ToPos:=i;
          if ToPos = FromPos then exit;

          pom:=FRecPoss[FromPos];                 //backup of object to move

          for i:=FromPos to ToPos-1 do
               FRecPoss[i]:=FRecPoss[i+1];
        end;

  FRecPoss[ToPos]:=pom;       //restoration of moved object
  FlushFRecPoss(FSavedAt);
end;

procedure TXDBTbl.XCHGRecord(FromPos, ToPos:integer);
var tmp:Int64;
begin
  FOwner.Owner.SetLastError(xdbrcInvalidIndex);

  if (FromPos<0) or (ToPos<0) or
     (FromPos > High(FRecPoss)) or (ToPos>High(FRecPoss)) then Abort;

  tmp:=FRecPoss[FromPos];
  FRecPoss[FromPos]:=FRecPoss[ToPos];
  FRecPoss[ToPos]:=tmp;
  FlushFRecPoss(FSavedAt);
  FOwner.Owner.SetLastError(xdbrcOK);
end;


end.
