//Part of XDB Core source code.
//Version 1.0
//Created by Tomas Koutny.
//for more info see http://www.rawos.cz/xdbe/index.html


//implementation of universal column of table
unit UCol;

interface

uses UserLevel, AbstractLayer, ColLevel, UserStructs;

type
   TUniversalColumn = class(TAXDBColumn) //read-only column multiplexor
                      protected
                        FSimulated:TXDBCol;
                        function GetID:Int64;
                        function GetOwner:TAXDBTable; override;
                        function FName:WideString;
                        function GetComment:WideString;
                        function GetInMask:WideString; override;
                        function GetOutMask:WideString; override;
                        function GetMinLength:Int64;
                        function GetMaxLength:Int64;
                        function GetMinVal:TXDBDataObject; override;
                        function GetMaxVal:TXDBDataObject; override;
                        function GetDefVal:TXDBDataObject; override;
                        procedure SetKind(NewKind:Int64); override;
                        function GetKind:Int64; override;
                      public
                        constructor Create(AOwner:TAXDBTable); override;
                        function VariableDataLength:boolean; override;
                        procedure Assign(NewDesign:TColumnDesignTX); override;
                        procedure Assign(ACol:TXDBColumn); override;
                                //change properties to new values


                        property Simulated:TXDBCol read FSimulated write FSimulated;

                        property ID:Int64 read GetID;
                        property Owner:TAXDBTable read GetOwner;

                        property Name:WideString read FName;
                        property Kind:Int64 read GetKind;
                        property Comment:WideString read GetComment;
                        property InputMask:WideString read GetInMask;
                        property OutputMask:WideString read GetOutMask;
                        property MinLength:Int64 read GetMinLength;
                        property MaxLength:Int64 read GetMaxLength;
                        property MinValue:TXDBDataObject read GetMinVal;
                        property MaxValue:TXDBDataObject read GetMaxVal;
                        property InitialValue:TXDBDataObject read GetDefVal;
                      end;

implementation

constructor TUniversalColumn.Create(AOwner:TAXDBTable);
begin
end;

function TUniversalColumn.GetID:Int64;
begin
  result:=FSimulated.ID;
end;

function TUniversalColumn.VariableDataLength:boolean;
begin
  result:=FSimulated.VariableDataLength;
end;

function TUniversalColumn.GetOwner:TAXDBTable;
begin
  result:=FSimulated.Owner;
end;

procedure TUniversalColumn.Assign(NewDesign:TColumnDesignTX);
begin
  FSimulated.Assign(NewDesign);
end;

procedure TUniversalColumn.Assign(ACol:TXDBColumn);
begin
  FSimulated.Assign(ACol);
end;

function TUniversalColumn.FName:WideString;
begin
  result:=FSimulated.Name;
end;

function TUniversalColumn.GetComment:WideString;
begin
  result:=FSimulated.Comment;
end;

function TUniversalColumn.GetInMask:WideString;
begin
  result:=FSimulated.InputMask;
end;

function TUniversalColumn.GetOutMask:WideString;
begin
  result:=FSimulated.OutputMask;
end;

function TUniversalColumn.GetMinLength:Int64;
begin
  result:=FSimulated.MinLength;
end;

function TUniversalColumn.GetMaxLength:Int64;
begin
  result:=FSimulated.MaxLength;
end;

function TUniversalColumn.GetMinVal:TXDBDataObject;
begin
  result:=FSimulated.MinValue;
end;

function TUniversalColumn.GetMaxVal:TXDBDataObject;
begin
  result:=FSimulated.MaxValue;
end;

function TUniversalColumn.GetDefVal:TXDBDataObject;
begin
  result:=FSimulated.InitialValue;
end;

procedure TUniversalColumn.SetKind(NewKind:Int64);
begin
  FSimulated.Kind:=NewKind;
end;

function TUniversalColumn.GetKind:Int64;
begin
  result:=FSimulated.Kind;
end;


end.
