//Part of XDBEngine source code.
//Version 1.0
//Created by Tomas Koutny.
//for more info see http://www.rawos.cz/xdbe/index.html

//definition of object representing one page in cache
unit CachePages;

interface

uses AbstractLayer, BTree, Stream64;

type
   TCPageNode = class;

   PCPageChain = ^TCPageChain;  //we will need to make some chain to get
   TCPageChain = record         //last recently used page quickly
                   prev, next:PCPageChain;
                   node:TCPageNode;
                 end;


   TCPageNode = class(TReferer) //we need to navigate through B Tree somehow
                public
                  Stream:TStream64;
                  Position:Int64;       //holds begin of the page in the file
                  FPage:PCachePage;     //points to memory that holds this data
                        //this variables will be set up by Main Engine
                  FChain:PCPageChain;   //we must have way how to update
                                        //LRU statistic

                  constructor Create;   //sets FreeOnRemove to false
                  destructor Destroy; override;  //frees FPage from memory
                        //doesn't alloc FPage because we want physical memory
                        //immediatelly and not when needed => FPage is allocated
                        //by main engine
                  function Compare(AReferer:TReferer):integer; override;
                end;

implementation

uses windows;

constructor TCPageNode.Create;
begin
  Stream:=nil;
  FPage:=nil;
  Position:=0;
  FreeOnRemove:=false;
end;

function TCPageNode.Compare(AReferer:TReferer):integer;
begin
                        //AReferer may be nil
                        //returns: -1 : self key < referer's key
                        //          0 : self key = referer's key
                        //          1 : self key > referer's key

  //our key is StreamPosition
  result:=1;
  if AReferer = nil then exit;

  if LongWord(Stream) = LongWord((AReferer as TCPageNode).Stream) then
     begin
       if Position = (AReferer as TCPageNode).Position then result:=0
         else if Position < (AReferer as TCPageNode).Position then result:=-1;
     end else
       if LongWord(Stream) < LongWord((AReferer as TCPageNode).Stream) then
         result:=-1;
end;

destructor TCPageNode.Destroy;
begin
  if FPage<>nil then
    begin       //if was not used then FPage is nil
      GlobalFree(cardinal(FPage^.Data));
      dispose(FPage);
      FPage:=nil;
    end;

  inherited;
end;

end.
